# -----------------------------------------------------------------------------
# Copyright (c) 2016- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
API to create an entry in Spyder Preferences associated to a given plugin.
"""

from __future__ import annotations

# Standard library imports
import types
import sys
from typing import TYPE_CHECKING

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias  # noqa: ICN003

# Local imports
from spyder.api.utils import PrefixedTuple
from spyder.config.manager import CONF
from spyder.config.types import ConfigurationKey
from spyder.widgets.config import SpyderConfigPage, BaseConfigTab

if TYPE_CHECKING:
    from qtpy.QtGui import QIcon

    from spyder.api.plugins import SpyderPluginV2
    import spyder.plugins.preferences.widget


OptionSet: TypeAlias = set[ConfigurationKey]
"""Type alias for a set of keys mapping to valid Spyder configuration values.

A :class:`set` of :class:`spyder.config.types.ConfigurationKey`\\s.
"""


class SpyderPreferencesTab(BaseConfigTab):
    """
    Widget that represents a tab on a preference page.

    All calls to :class:`spyder.widgets.config.SpyderConfigPage` attributes
    are resolved via delegation.
    """

    TITLE: str | None = None
    """Name of the tab to display; must be set on the child implementations."""

    def __init__(self, parent: SpyderConfigPage) -> None:
        """
        Create a new tab on the given ``parent`` config page.

        Parameters
        ----------
        parent : spyder.widgets.config.SpyderConfigPage
            The config page the tab will live on, to be this widget's parent.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If :attr:`TITLE` has not been set on the child implementation.
        """
        super().__init__(parent)
        self.parent = parent

        if self.TITLE is None or not isinstance(self.TITLE, str):
            raise ValueError("TITLE must be a str")

    def apply_settings(self) -> OptionSet:
        """
        Hook to manually apply settings that cannot be applied automatically.

        Reimplement this if the configuration tab has complex widgets that
        cannot be created with any of the ``self.create_*`` calls.
        This call should return a :class:`set` containing the configuration
        options that changed.

        Returns
        -------
        OptionSet
            The :class:`set` of Spyder
            :class:`~spyder.config.types.ConfigurationKey`\\s that were
            manually applied.
        """
        return set({})

    def is_valid(self) -> bool:
        """
        Return ``False`` if the tab contents are invalid, ``True`` otherwise.

        This method can be overriden to perform complex checks.

        Returns
        -------
        bool
            Whether the tab contents are valid.
        """
        return True

    def __getattr__(self, attr):
        this_class_dir = dir(self)
        if attr not in this_class_dir:
            return getattr(self.parent, attr)
        else:
            return super().__getattr__(attr)

    def setLayout(self, layout):
        """Remove default margins around the layout by default."""
        layout.setContentsMargins(0, 0, 0, 0)
        super().setLayout(layout)


class PluginConfigPage(SpyderConfigPage):
    """
    A Spyder :guilabel:`Preferences` page for a single plugin.

    This widget exposes the options a plugin offers for configuration as
    an entry in Spyder's :guilabel:`Preferences` dialog.
    """

    # TODO: Temporal attribute to handle which appy_settings method to use
    # the one of the conf page or the one in the plugin, while the config
    # dialog system is updated.
    APPLY_CONF_PAGE_SETTINGS = False

    def __init__(
        self,
        plugin: SpyderPluginV2,
        parent: spyder.plugins.preferences.widget.ConfigDialog,
    ) -> None:
        """
        Create a Spyder :guilabel:`Preferences` page for a plugin.

        Parameters
        ----------
        plugin : SpyderPluginV2
            The plugin to create the configuration page for.
        parent : spyder.plugins.preferences.widgets.configdialog.ConfigDialog
            The main Spyder :guilabel:`Preferences` dialog, parent widget
            to this one.

        Returns
        -------
        None
        """
        self.plugin = plugin
        self.main = parent.main

        if hasattr(plugin, "CONF_SECTION"):
            self.CONF_SECTION = plugin.CONF_SECTION

        if hasattr(plugin, "get_font"):
            self.get_font = plugin.get_font

        if not self.APPLY_CONF_PAGE_SETTINGS:
            self._patch_apply_settings(plugin)

        SpyderConfigPage.__init__(self, parent)

    def _wrap_apply_settings(self, func):
        """
        Wrap apply_settings call to ensure that a user-defined custom call
        is called alongside the Spyder Plugin API configuration propagation
        call.
        """

        def wrapper(self, options):
            opts = self.previous_apply_settings() or set({})
            opts |= options
            self.aggregate_sections_partials(opts)
            func(opts)

        return types.MethodType(wrapper, self)

    def _patch_apply_settings(self, plugin):
        self.previous_apply_settings = self.apply_settings
        self.apply_settings = self._wrap_apply_settings(plugin.apply_conf)
        self.get_option = plugin.get_conf
        self.set_option = plugin.set_conf
        self.remove_option = plugin.remove_conf

    def aggregate_sections_partials(self, opts: OptionSet) -> None:
        """
        Aggregate options by sections in order to notify observers.

        Parameters
        ----------
        opts : OptionSet
            The options to aggregate by section.

        Returns
        -------
        None
        """
        to_update = {}
        for opt in opts:
            if isinstance(opt, tuple):
                # This is necessary to filter tuple options that do not
                # belong to a section.
                if len(opt) == 2 and opt[0] is None:
                    opt = opt[1]

            section = self.CONF_SECTION
            if opt in self.cross_section_options:
                section = self.cross_section_options[opt]
            section_options = to_update.get(section, [])
            section_options.append(opt)
            to_update[section] = section_options

        for section in to_update:
            section_prefix = PrefixedTuple()
            # Notify section observers
            CONF.notify_observers(
                section, "__section", recursive_notification=False
            )
            for opt in to_update[section]:
                if isinstance(opt, tuple):
                    opt = opt[:-1]
                    section_prefix.add_path(opt)
            # Notify prefixed observers
            for prefix in section_prefix:
                try:
                    CONF.notify_observers(
                        section, prefix, recursive_notification=False
                    )
                except Exception:
                    # Prevent unexpected failures on tests
                    pass

    def get_name(self) -> str:
        """
        Return plugin name to use in preferences page title and message boxes.

        Normally you do not have to reimplement it, as the plugin name
        in the preferences page will be the same as the plugin title.

        Returns
        -------
        str
            The plugin this preference page corresponds to.
        """
        return self.plugin.get_name()

    def get_icon(self) -> QIcon:
        """
        Return the plugin icon to use in the preferences page.

        Normally you do not have to reimplement it, as the plugin icon
        in the preferences page will be the same as the main plugin icon.

        Returns
        -------
        QIcon
            The plugin's icon.
        """
        return self.plugin.get_icon()

    def setup_page(self) -> None:
        """
        Set up the configuration page widget.

        You need to implement this method to set the layout of the
        preferences page.

        For example:

        .. code-block:: python

            layout = QVBoxLayout()
            layout.addWidget(...)
            ...
            self.setLayout(layout)

        Returns
        -------
        None
        """
        raise NotImplementedError

    def apply_settings(self) -> OptionSet:
        """
        Hook to manually apply settings that cannot be applied automatically.

        Reimplement this if the configuration tab has complex widgets that
        cannot be created with any of the ``self.create_*`` calls.
        This call should return a :class:`set` containing the configuration
        options that changed.

        Returns
        -------
        OptionSet
            The :class:`set` of Spyder
            :class:`~spyder.config.types.ConfigurationKey`\\s that were
            manually applied.
        """
        return set({})



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import asyncio
import fnmatch
import io
import logging
import os
import posixpath
from datetime import datetime

from qtpy.compat import getopenfilename, getsavefilename
from qtpy.QtCore import QSortFilterProxyModel, Qt, Signal
from qtpy.QtGui import QClipboard, QStandardItem, QStandardItemModel
from qtpy.QtWidgets import (
    QApplication,
    QInputDialog,
    QMessageBox,
    QLineEdit,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.config.decorators import on_conf_change
from spyder.api.translations import _
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import get_conf_path
from spyder.config.utils import EDIT_EXTENSIONS
from spyder.plugins.editor.utils.editor import get_default_file_content
from spyder.plugins.remoteclient.api.modules.base import (
    SpyderRemoteSessionClosed,
)
from spyder.plugins.remoteclient.api.modules.file_services import (
    RemoteFileServicesError,
    RemoteOSError,
)
from spyder.utils.icon_manager import ima
from spyder.utils.misc import getcwd_or_home


logger = logging.getLogger(__name__)


class RemoteViewMenus:
    Context = "remote_context_menu"
    New = "remote_new_menu"


class RemoteViewNewSubMenuSections:
    General = 'remote_general_section'
    Language = 'remote_language_section'


class RemoteExplorerActions:
    CopyPaste = "remote_copy_paste_action"
    CopyPath = "remote_copy_path_action"
    Delete = "remote_delete_action"
    Download = "remote_download_action"
    NewDirectory = "remote_new_directory_action"
    NewFile = "remote_new_file_action"
    NewModule = "remote_new_module_action"
    NewPackage = "remote_new_package_action"
    Rename = "remote_rename_action"
    Upload = "remote_upload_file_action"


class RemoteExplorerContextMenuSections:
    CopyPaste = "remote_copy_paste_section"
    Extras = "remote_extras_section"
    New = "remote_new_section"


class RemoteQSortFilterProxyModel(QSortFilterProxyModel):

    def lessThan(self, left, right):
        right_data = self.sourceModel().data(
            self.sourceModel().index(right.row(), 0), Qt.UserRole + 1
        )
        if right_data["type"] == "ACTION":
            return self.sortOrder() == Qt.AscendingOrder

        left_data = self.sourceModel().data(
            self.sourceModel().index(left.row(), 0), Qt.UserRole + 1
        )
        if left_data["type"] == "ACTION":
            return self.sortOrder() == Qt.DescendingOrder

        if left_data["type"] == "directory" and right_data["type"] == "file":
            return True

        if left_data["type"] == "file" and right_data["type"] == "directory":
            return False

        return super().lessThan(left, right)


class RemoteExplorer(QWidget, SpyderWidgetMixin):
    sig_dir_opened = Signal(str, str)
    sig_start_spinner_requested = Signal()
    sig_stop_spinner_requested = Signal()

    def __init__(self, parent=None, class_parent=None, files=None):
        super().__init__(parent=parent, class_parent=parent)

        # General attributes
        self.remote_files_manager = None
        self.server_id: str | None = None
        self.root_prefix: dict[str, str] = {}

        self.background_files_load = set()
        self.extra_files = []
        self.more_files_available = False

        self.filter_on = False
        self.name_filters = []

        self.history = []
        self.histindex = None

        # Model, actions and widget setup
        self.context_menu = self.create_menu(RemoteViewMenus.Context)
        new_submenu = self.create_menu(
            RemoteViewMenus.New,
            _('New'),
        )

        self.new_package_action = self.create_action(
            RemoteExplorerActions.NewPackage,
            text=_("Python package..."),
            icon=self.create_icon('package_new'),
            triggered=self.new_package,
        )
        self.new_module_action = self.create_action(
            RemoteExplorerActions.NewModule,
            text=_("Python file..."),
            icon=self.create_icon('python'),
            triggered=self.new_module,
        )
        self.new_directory_action = self.create_action(
            RemoteExplorerActions.NewDirectory,
            text=_("Folder..."),
            icon=self.create_icon('folder_new'),
            triggered=self.new_directory,
        )
        self.new_file_action = self.create_action(
            RemoteExplorerActions.NewFile,
            text=_("File..."),
            icon=self.create_icon('TextFileIcon'),
            triggered=self.new_file,
        )
        self.copy_paste_action = self.create_action(
            RemoteExplorerActions.CopyPaste,
            _("Copy and Paste..."),
            icon=self.create_icon("editcopy"),
            triggered=self.copy_paste_item,
        )
        self.rename_action = self.create_action(
            RemoteExplorerActions.Rename,
            _("Rename..."),
            icon=self.create_icon("rename"),
            triggered=self.rename_item,
        )
        self.copy_path_action = self.create_action(
            RemoteExplorerActions.CopyPath,
            _("Copy path"),
            triggered=self.copy_path,
        )
        self.delete_action = self.create_action(
            RemoteExplorerActions.Delete,
            _("Delete..."),
            icon=self.create_icon("editclear"),
            triggered=self.delete_item,
        )
        self.download_action = self.create_action(
            RemoteExplorerActions.Download,
            _("Download..."),
            icon=self.create_icon("fileimport"),
            triggered=self.download_item,
        )
        self.upload_file_action = self.create_action(
            RemoteExplorerActions.Upload,
            _("Upload file"),
            icon=self.create_icon("fileexport"),
            triggered=self.upload_file,
        )

        for item in [self.new_file_action, self.new_directory_action]:
            self.add_item_to_menu(
                item,
                new_submenu,
                section=RemoteViewNewSubMenuSections.General,
            )

        for item in [self.new_module_action, self.new_package_action]:
            self.add_item_to_menu(
                item,
                new_submenu,
                section=RemoteViewNewSubMenuSections.Language,
            )

        for item in [
            new_submenu,
            self.rename_action,
            self.delete_action,
        ]:
            self.add_item_to_menu(
                item,
                self.context_menu,
                section=RemoteExplorerContextMenuSections.New,
            )

        for item in [
            self.copy_paste_action,
            self.copy_path_action,
        ]:
            self.add_item_to_menu(
                item,
                self.context_menu,
                section=RemoteExplorerContextMenuSections.CopyPaste,
            )

        for item in [self.download_action]:
            self.add_item_to_menu(
                item,
                self.context_menu,
                section=RemoteExplorerContextMenuSections.Extras,
            )

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_show_context_menu)

        self.model = QStandardItemModel(self)
        self.model.setHorizontalHeaderLabels(
            [
                "Name",
                "Size",
                "Type",
                "Date Modified",
            ]
        )
        self.proxy_model = RemoteQSortFilterProxyModel(self)
        self.proxy_model.setSourceModel(self.model)

        self.view = QTreeView(self)
        self.view.setModel(self.proxy_model)
        self.view.setSortingEnabled(True)
        self.view.setContextMenuPolicy(Qt.CustomContextMenu)
        self.view.customContextMenuRequested.connect(
            self._on_show_context_menu
        )

        if files:
            self.set_files(files)

        self.view.sortByColumn(0, Qt.AscendingOrder)
        self.view.entered.connect(self._on_entered_item)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.view)

    @on_conf_change(
        option=[
            "size_column",
            "type_column",
            "date_column",
            "name_filters",
            "show_hidden",
            "single_click_to_open",
        ]
    )
    def on_conf_update(self, option, value):
        if option == "size_column":
            self.view.setColumnHidden(1, not value)
        elif option == "type_column":
            self.view.setColumnHidden(2, not value)
        elif option == "date_column":
            self.view.setColumnHidden(3, not value)
        elif option == "name_filters":
            if self.filter_on:
                self.filter_files(value)
        elif option == "show_hidden":
            self.refresh(force_current=True)
        elif option == "single_click_to_open":
            self.set_single_click_to_open(value)

    def _on_show_context_menu(self, position):
        index = self.view.indexAt(position)
        if not index.isValid():
            self.view.setCurrentIndex(index)
            self.view.clearSelection()
            data = {}
            data_available = False
            is_file = False
        else:
            source_index = self.proxy_model.mapToSource(
                self.view.currentIndex()
            )
            data_index = self.model.index(source_index.row(), 0)
            data = self.model.data(data_index, Qt.UserRole + 1)
            data_available = bool(data)
            is_file = (
                data_available and data.get("type", "directory") == "file"
            )

        # Disable actions that require a valid selection
        self.rename_action.setEnabled(data_available)
        self.delete_action.setEnabled(data_available)

        # Disable actions not suitable for directories
        self.copy_paste_action.setEnabled(is_file)

        global_position = self.mapToGlobal(position)
        self.context_menu.popup(global_position)

    def _on_entered_item(self, index):
        if self.get_conf("single_click_to_open"):
            self.view.setCursor(Qt.PointingHandCursor)
            self.view.header().setCursor(Qt.ArrowCursor)
        else:
            self.view.setCursor(Qt.ArrowCursor)

    def _on_clicked_item(self, index):
        source_index = self.proxy_model.mapToSource(index)
        data_index = self.model.index(source_index.row(), 0)
        data = self.model.data(data_index, Qt.UserRole + 1)
        if data:
            data_name = data["name"]
            data_type = data["type"]
            if data_type == "directory":
                self.chdir(data_name, emit=True)
            elif (
                data_type == "file"
                and os.path.splitext(data_name)[1] in EDIT_EXTENSIONS
            ):
                QMessageBox.warning(
                    self,
                    _("Unsupported functionality"),
                    _(
                        "Opening remote files in the Editor is not yet "
                        "supported. This feature will be added in Spyder "
                        "6.2.0"
                    )
                )
            elif data_type == "ACTION" and data_name == "FETCH_MORE":
                self.fetch_more_files()

    def _handle_future_response_error(self, response, error_title, error_message):
        result = response.result()
        if isinstance(result, Exception):
            logger.debug(result)
            error_message = f"{error_message}<br><br>{result.message}"
            QMessageBox.critical(self, error_title, error_message)

    @AsyncDispatcher.QtSlot
    def _on_remote_new_package(self, future, package_name):
        self._handle_future_response_error(
            future,
            _("New Python Package error"),
            _("An error occured while trying to create a new Python package"),
        )
        new_name = posixpath.join(package_name, "__init__.py")
        self._new_item(new_name, for_file=True, with_content=True)

    @AsyncDispatcher.QtSlot
    def _on_remote_new_module(self, future):
        self._handle_future_response_error(
            future,
            _("New Python File error"),
            _("An error occured while trying to create a new Python file"),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_new_module(self, new_path, file_content):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        try:
            file_manager = await self.remote_files_manager.open(new_path, mode="w")
            remote_file_response = await file_manager.write(file_content)
            await file_manager.close()
        except (RemoteOSError, OSError) as error:
            # Catch error raised when awaiting. The error will be available
            # and will be shown when handling the result with the usage of
            # `_handle_future_response_error`
            # See spyder-ide/spyder#24974
            remote_file_response = error

        return remote_file_response

    @AsyncDispatcher.QtSlot
    def _on_remote_new(self, future):
        self._handle_future_response_error(
            future,
            _("New error"),
            _("An error occured while trying to create a file/directory"),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_new(
        self, new_path, for_file=False
    ):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        try:
            if for_file:
                response = await self.remote_files_manager.touch(new_path)
            else:
                response = await self.remote_files_manager.mkdir(new_path)
        except (RemoteOSError, OSError) as error:
            # Catch error raised when awaiting. The error will be available
            # and will be shown when handling the result with the usage of
            # `_handle_future_response_error`
            # See spyder-ide/spyder#24974
            response = error

        return response

    @AsyncDispatcher.QtSlot
    def _on_remote_copy_paste(self, future):
        self._handle_future_response_error(
            future,
            _("Copy and Paste error"),
            _(
                "An error occured while trying to copy and paste a "
                "file/directory"
            ),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_copy_paste(self, old_path, new_path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        return await self.remote_files_manager.copy(old_path, new_path)

    @AsyncDispatcher.QtSlot
    def _on_remote_rename(self, future):
        self._handle_future_response_error(
            future,
            _("Rename error"),
            _("An error occured while trying to rename a file"),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_rename(self, old_path, new_path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        return await self.remote_files_manager.replace(old_path, new_path)

    @AsyncDispatcher.QtSlot
    def _on_remote_delete(self, future):
        self._handle_future_response_error(
            future,
            _("Delete error"),
            _("An error occured while trying to delete a file/directory"),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_delete(self, path, is_file=False):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        try:
            if is_file:
                response = await self.remote_files_manager.unlink(path)
            else:
                response = await self.remote_files_manager.rmdir(
                    path, non_empty=True
                )
        except (RemoteOSError, OSError) as error:
            # Catch error raised when awaiting. The error will be available
            # and will be shown when handling the result with the usage of
            # `_handle_future_response_error`
            # See spyder-ide/spyder#24974
            response = error

        return response

    @AsyncDispatcher.QtSlot
    def _on_remote_download_file(self, future, remote_filename):
        data = future.result()
        filename, __ = getsavefilename(
            self,
            _("Download file"),
            os.path.join(getcwd_or_home(), remote_filename),
            _("All files") + " (*)",
        )
        if filename:
            try:
                with open(filename, "w") as download_file:
                    download_file.write(data)
            except TypeError:
                with open(filename, "wb") as download_file:
                    download_file.write(data)

        self.sig_stop_spinner_requested.emit()

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_download_directory(self, path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        try:
            zip_generator = self.remote_files_manager.zip_directory(path)
            zip_data = io.BytesIO()
            async for data in zip_generator:
                zip_data.write(data)
            zip_data.seek(0)
        except RemoteFileServicesError as download_error:
            logger.debug(f"Unable to download {path}")
            logger.debug(
                f"Error while trying to download directory (compressed): "
                f"{download_error.message}"
            )
            error_message = _(
                "An error occured while trying to download {path}"
            ).format(path=path)
            QMessageBox.critical(self, _("Download error"), error_message)

        return zip_data.getbuffer()

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_download_file(self, path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        try:
            file_manager = await self.remote_files_manager.open(path, mode="r")
            file_data = ""
            async for data in file_manager:
                file_data += data
        except RemoteFileServicesError:
            try:
                file_manager = await self.remote_files_manager.open(
                    path, mode="rb"
                )
                file_data = b""
                async for data in file_manager:
                    file_data += data
            except RemoteFileServicesError as download_error:
                logger.debug(f"Unable to download {path}")
                logger.debug(
                    f"Error while trying to download file: "
                    f"{download_error.message}"
                )
                error_message = _(
                    "An error occured while trying to download {path}"
                ).format(path=path)
                QMessageBox.critical(self, _("Download error"), error_message)

        await file_manager.close()
        return file_data

    @AsyncDispatcher.QtSlot
    def _on_remote_upload_file(self, future):
        self._handle_future_response_error(
            future,
            _("Upload error"),
            _("An error occured while trying to upload a file"),
        )
        self.refresh(force_current=True)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_upload_file(self, local_path):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        remote_file = posixpath.join(
            self.root_prefix[self.server_id], os.path.basename(local_path)
        )
        file_content = None

        try:
            try:
                with open(local_path, mode="r") as local_file:
                    file_content = local_file.read()
                file_manager = await self.remote_files_manager.open(
                    remote_file, mode="w"
                )
            except UnicodeDecodeError:
                with open(local_path, mode="rb") as local_file:
                    file_content = local_file.read()
                file_manager = await self.remote_files_manager.open(
                    remote_file, mode="wb"
                )

            if file_content:
                remote_file_response = await file_manager.write(file_content)

            await file_manager.close()
        except (RemoteOSError, OSError) as error:
            # Catch error raised when awaiting. The error will be available
            # and will be shown when handling the result with the usage of
            # `_handle_future_response_error`
            # See spyder-ide/spyder#24974
            remote_file_response = error

        return remote_file_response

    @AsyncDispatcher.QtSlot
    def _on_remote_ls(self, future):
        data = future.result()
        self.set_files(data)

    @AsyncDispatcher(loop="explorer")
    async def _do_remote_ls(self, path, server_id):
        if not self.remote_files_manager:
            self.sig_stop_spinner_requested.emit()
            return

        for task in self.background_files_load:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

        self.extra_files = []
        self.more_files_available = False
        files = []
        try:
            init_files_display = self.get_conf("init_files_display")
            generator = self.remote_files_manager.ls(path)
            async for file in generator:
                file_name = os.path.relpath(
                    file["name"], self.root_prefix[self.server_id]
                )
                file_type = file["type"]
                if len(files) < init_files_display:
                    if not self.get_conf(
                        "show_hidden"
                    ) and file_name.startswith("."):
                        continue
                    if (
                        self.name_filters
                        and len(self.name_filters)
                        and file_type == "file"
                    ):
                        for name_filter in self.name_filters:
                            if file_type == "file" and fnmatch.fnmatch(
                                file_name, name_filter
                            ):
                                files.append(file)
                                break
                    else:
                        files.append(file)
                else:
                    break

            if len(files) == init_files_display:
                task = asyncio.create_task(
                    self._get_extra_files(generator, len(files))
                )
                self.background_files_load.add(task)
                task.add_done_callback(self.background_files_load.discard)

        except RemoteOSError as error:
            # TODO: Should the error be shown in some way?
            logger.debug(error)
        except SpyderRemoteSessionClosed:
            self.remote_files_manager = None

        return files

    async def _get_extra_files(self, generator, already_added):
        self.extra_files = []
        self.more_files_available = False
        async for file in generator:
            if len(self.extra_files) + already_added == self.get_conf(
                "max_files_display"
            ):
                self.more_files_available = True
                break

            file_name = os.path.relpath(
                file["name"], self.root_prefix[self.server_id]
            )
            file_type = file["type"]
            if not self.get_conf("show_hidden") and file_name.startswith("."):
                continue

            if (
                self.name_filters
                and len(self.name_filters)
                and file_type == "file"
            ):
                for name_filter in self.name_filters:
                    if file_type == "file" and fnmatch.fnmatch(
                        file_name, name_filter
                    ):
                        self.extra_files.append(file)
                        break
            else:
                self.extra_files.append(file)

        logger.debug(
            f"{len(self.extra_files)} available extra files to be shown"
        )

    def _new_item(self, new_name, for_file=False, with_content=False):
        new_path = posixpath.join(self.root_prefix[self.server_id], new_name)
        if not with_content:
            self._do_remote_new(new_path, for_file=for_file).connect(
                self._on_remote_new
            )
        elif for_file and with_content:
            file_content, __, __ = get_default_file_content(
                get_conf_path("template.py")
            )
            self._do_remote_new_module(new_path, file_content).connect(
                self._on_remote_new_module
            )
        elif not for_file and with_content:

            @AsyncDispatcher.QtSlot
            def remote_package(future):
                self._on_remote_new_package(future, new_name)

            self._do_remote_new(new_path).connect(remote_package)

        self.sig_start_spinner_requested.emit()

    @AsyncDispatcher.QtSlot
    def chdir(
        self,
        directory=None,
        server_id=None,
        emit=True,
        browsing_history=False,
        remote_files_manager=None,
    ):
        if browsing_history:
            directory = self.history[self.histindex]
        elif directory in self.history:
            self.histindex = self.history.index(directory)
        else:
            if self.histindex is None:
                self.history = []
            else:
                self.history = self.history[: self.histindex + 1]
            if len(self.history) == 0 or (
                self.history and self.history[-1] != directory
            ):
                self.history.append(directory)
            self.histindex = len(self.history) - 1

        if directory == self.root_prefix.get(server_id):
            return

        if server_id:
            self.server_id = server_id
        self.root_prefix[self.server_id] = directory
        if remote_files_manager:
            self.remote_files_manager = remote_files_manager
        self.refresh(force_current=True)
        if emit:
            self.sig_dir_opened.emit(directory, self.server_id)

    def set_files(self, files, reset=True):
        if reset:
            self.model.setRowCount(0)

        if files:
            logger.debug(f"Setting {len(files)} files")
            root = self.model.invisibleRootItem()

            more_files_items = self.model.match(
                self.model.index(0, 0), Qt.DisplayRole, _("Show more files")
            )
            if len(more_files_items):
                # Remove more files item
                self.model.removeRow(more_files_items[-1].row())

            more_files_available = self.model.match(
                self.model.index(0, 0),
                Qt.DisplayRole,
                _("Maximum number of files to display reached!"),
            )
            if len(more_files_available):
                # Remove more items available item
                self.model.removeRow(more_files_available[-1].row())

            for file in files:
                path = file["name"]
                name = os.path.relpath(path, self.root_prefix[self.server_id])

                file_type = file["type"]
                icon = ima.icon("FileIcon")
                if file_type == "directory":
                    icon = ima.icon("DirClosedIcon")

                file_name = QStandardItem(icon, name)
                file_name.setData(file)
                file_name.setToolTip(file["name"])

                file_size = QStandardItem(str(file["size"]))
                file_type = QStandardItem(file_type)
                file_date_modified = QStandardItem(
                    datetime.fromtimestamp(file["mtime"]).strftime(
                        "%d/%m/%Y %I:%M %p"
                    )
                )

                items = [
                    file_name,
                    file_size,
                    file_type,
                    file_date_modified,
                ]
                for standard_item in items:
                    standard_item.setEditable(False)
                root.appendRow(items)

            # Add fetch more or more items available item
            if len(self.extra_files):
                fetch_more_item = QStandardItem(_("Show more files"))
                fetch_more_item.setEditable(False)
                fetch_more_item.setData(
                    {"name": "FETCH_MORE", "type": "ACTION"}
                )
                root.appendRow(fetch_more_item)
                self.view.setFirstColumnSpanned(
                    fetch_more_item.index().row(), root.index(), True
                )
            elif len(self.extra_files) == 0 and self.more_files_available:
                more_items_available = QStandardItem(
                    _("Maximum number of files to display reached!")
                )
                more_items_available.setEditable(False)
                more_items_available.setData(
                    {"name": "MESSAGE", "type": "ACTION"}
                )
                root.appendRow(more_items_available)
                self.view.setFirstColumnSpanned(
                    more_items_available.index().row(), root.index(), True
                )

            self.view.resizeColumnToContents(0)

    def fetch_more_files(self):
        fetch_files_display = self.get_conf("fetch_files_display")
        new_files = self.extra_files[:fetch_files_display]
        del self.extra_files[:fetch_files_display]
        self.set_files(new_files, reset=False)
        logger.debug(
            f"{len(self.extra_files)} extra files remaining to be shown"
        )

    def set_current_folder(self, folder):
        self.root_prefix[self.server_id] = folder
        return self.model.invisibleRootItem()

    def get_current_folder(self):
        return self.root_prefix[self.server_id]

    def go_to_parent_directory(self):
        parent_directory = os.path.dirname(self.root_prefix[self.server_id])
        logger.debug(
            f"Going to parent directory of {self.root_prefix[self.server_id]}: "
            f"{parent_directory}"
        )
        self.chdir(parent_directory)

    def go_to_previous_directory(self):
        self.histindex -= 1
        logger.debug(
            f"Going to previous directory in history with index "
            f"{self.histindex}"
        )
        self.chdir(browsing_history=True)

    def go_to_next_directory(self):
        self.histindex += 1
        logger.debug(
            f"Going to next directory in history with index {self.histindex}"
        )
        self.chdir(browsing_history=True)

    def refresh(self, new_path=None, force_current=False):
        if force_current:
            if new_path is None:
                new_path = self.root_prefix.get(self.server_id)
            self._do_remote_ls(new_path, self.server_id).connect(
                self._on_remote_ls
            )

        self.previous_action.setEnabled(False)
        self.next_action.setEnabled(False)

        if self.histindex is not None:
            self.previous_action.setEnabled(self.histindex > 0)
            self.next_action.setEnabled(self.histindex < len(self.history) - 1)
        self.sig_stop_spinner_requested.emit()

    def set_single_click_to_open(self, value):
        if value:
            try:
                self.view.doubleClicked.disconnect(self._on_clicked_item)
            except TypeError:
                pass
            self.view.clicked.connect(self._on_clicked_item)
        else:
            try:
                self.view.clicked.disconnect(self._on_clicked_item)
            except TypeError:
                pass
            self.view.doubleClicked.connect(self._on_clicked_item)

    def filter_files(self, name_filters=None):
        """Filter files given the defined list of filters."""
        if name_filters is None:
            name_filters = self.get_conf("name_filters")
        self.name_filters = []
        if self.filter_on:
            self.name_filters = name_filters
        self.refresh(force_current=True)

    def change_filter_state(self):
        self.filter_on = not self.filter_on
        self.filter_button.setChecked(self.filter_on)
        self.filter_button.setToolTip(_("Filter filenames"))
        self.filter_files()

    def new_package(self):
        new_name, valid = QInputDialog.getText(
            self, _("New Python Package"), _("Name as:"), QLineEdit.Normal, ""
        )
        if valid:
            self._new_item(new_name, with_content=True)

    def new_module(self):
        new_name, valid = QInputDialog.getText(
            self, _("New Python File"), _("Name as:"), QLineEdit.Normal, ".py"
        )
        if valid:
            self._new_item(new_name, for_file=True, with_content=True)

    def new_directory(self):
        new_name, valid = QInputDialog.getText(
            self, _("New Folder"), _("Name as:"), QLineEdit.Normal, ""
        )
        if valid:
            self._new_item(new_name)

    def new_file(self):
        new_name, valid = QInputDialog.getText(
            self, _("New File"), _("Name as:"), QLineEdit.Normal, ""
        )
        if valid:
            self._new_item(new_name, for_file=True)

    def copy_paste_item(self):
        if (
            not self.view.currentIndex()
            or not self.view.currentIndex().isValid()
        ):
            return

        source_index = self.proxy_model.mapToSource(self.view.currentIndex())
        data_index = self.model.index(source_index.row(), 0)
        data = self.model.data(data_index, Qt.UserRole + 1)
        if data:
            old_path = data["name"]
            relpath = os.path.relpath(
                old_path, self.root_prefix[self.server_id]
            )
            new_relpath, valid = QInputDialog.getText(
                self,
                _("Copy and Paste"),
                _("Paste as:"),
                QLineEdit.Normal,
                relpath,
            )
            if valid:
                new_path = posixpath.join(
                    self.root_prefix[self.server_id], new_relpath
                )
                self._do_remote_copy_paste(old_path, new_path).connect(
                    self._on_remote_copy_paste
                )
                self.sig_start_spinner_requested.emit()

    def rename_item(self):
        if (
            not self.view.currentIndex()
            or not self.view.currentIndex().isValid()
        ):
            return

        source_index = self.proxy_model.mapToSource(self.view.currentIndex())
        data_index = self.model.index(source_index.row(), 0)
        data = self.model.data(data_index, Qt.UserRole + 1)
        if data:
            old_path = data["name"]
            relpath = os.path.relpath(
                old_path, self.root_prefix[self.server_id]
            )
            new_relpath, valid = QInputDialog.getText(
                self, _("Rename"), _("New name:"), QLineEdit.Normal, relpath
            )
            if valid:
                new_path = posixpath.join(
                    self.root_prefix[self.server_id], new_relpath
                )
                self._do_remote_rename(old_path, str(new_path)).connect(
                    self._on_remote_rename
                )
                self.sig_start_spinner_requested.emit()

    def copy_path(self):
        if (
            not self.view.currentIndex()
            or not self.view.currentIndex().isValid()
        ):
            path = self.root_prefix[self.server_id]
        else:
            source_index = self.proxy_model.mapToSource(
                self.view.currentIndex()
            )
            data_index = self.model.index(source_index.row(), 0)
            data = self.model.data(data_index, Qt.UserRole + 1)
            if data:
                path = data["name"]

        cb = QApplication.clipboard()
        cb.setText(path, mode=QClipboard.Mode.Clipboard)

    def delete_item(self):
        if (
            not self.view.currentIndex()
            or not self.view.currentIndex().isValid()
        ):
            return

        source_index = self.proxy_model.mapToSource(self.view.currentIndex())
        data_index = self.model.index(source_index.row(), 0)
        data = self.model.data(data_index, Qt.UserRole + 1)
        if data:
            path = data["name"]
            filename = os.path.relpath(path, self.root_prefix[self.server_id])
            result = QMessageBox.warning(
                self,
                _("Delete"),
                _("Do you really want to delete <b>{filename}</b>?").format(
                    filename=filename
                ),
                QMessageBox.Yes | QMessageBox.No,
            )
            if result == QMessageBox.Yes:
                is_file = data["type"] == "file"
                self._do_remote_delete(path, is_file=is_file).connect(
                    self._on_remote_delete
                )
                self.sig_start_spinner_requested.emit()

    def download_item(self):
        if (
            not self.view.currentIndex()
            or not self.view.currentIndex().isValid()
        ):
            return

        source_index = self.proxy_model.mapToSource(self.view.currentIndex())
        data_index = self.model.index(source_index.row(), 0)
        data = self.model.data(data_index, Qt.UserRole + 1)
        if data:
            path = data["name"]
            filename = os.path.relpath(path, self.root_prefix[self.server_id])
            is_file = data["type"] == "file"
            remote_filename = filename if is_file else f"{filename}.zip"

            @AsyncDispatcher.QtSlot
            def remote_download_file(future):
                self._on_remote_download_file(future, remote_filename)

            if is_file:
                self._do_remote_download_file(path).connect(
                    remote_download_file
                )
            else:
                self._do_remote_download_directory(path).connect(
                    remote_download_file
                )

            self.sig_start_spinner_requested.emit()

    def upload_file(self):
        local_path, __ = getopenfilename(
            self,
            _("Upload file"),
            getcwd_or_home(),
            _("All files") + " (*)",
        )
        if os.path.exists(local_path):
            self._do_remote_upload_file(local_path).connect(
                self._on_remote_upload_file
            )
            self.sig_start_spinner_requested.emit()

    def reset(self, server_id):
        self.root_prefix[server_id] = None



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
spyder.plugins.remoteclient.api.manager
=======================================

Remote Client Plugin API Manager.
"""

from __future__ import annotations

import asyncio
import json

import aiohttp
import yarl
from packaging.version import Version

from spyder.api.translations import _
from spyder.plugins.remoteclient import (
    SPYDER_REMOTE_MAX_VERSION,
    SPYDER_REMOTE_MIN_VERSION,
    SPYDER_PLUGIN_NAME
)
from spyder.plugins.remoteclient.api.manager.base import (
    SpyderRemoteAPIManagerBase,
)
from spyder.plugins.remoteclient.api.protocol import (
    ConnectionStatus,
    JupyterHubClientOptions,
)


class SpyderRemoteJupyterHubAPIManager(SpyderRemoteAPIManagerBase):
    """Class to manage a remote JupyterHub server and its APIs."""

    def __init__(
        self, conf_id, options: JupyterHubClientOptions, _plugin=None
    ):
        super().__init__(conf_id, options, _plugin)

        self._server_url = None
        self._user_name = None
        self._session: aiohttp.ClientSession = None
        self._hb_task: asyncio.Task = None

    @property
    def api_token(self):
        return self.options["token"]

    @property
    def server_url(self):
        return self.hub_url / f"user/{self._user_name}/spyder/"

    @property
    def hub_url(self):
        return yarl.URL(self.options["url"])

    async def _start_remote_server(self):
        """Start remote server."""
        if not self.connected:
            self.logger.error("SSH connection is not open")
            self._emit_connection_status(
                ConnectionStatus.Error,
                _("JupyterHub is not connected"),
            )
            return False

        async with self._session.post(
            f"hub/api/users/{self._user_name}/servers/spyder",
        ) as response:
            if response.status in {201, 400}:
                if await self.check_server_version():
                    self._hb_task = asyncio.create_task(self._heartbeat())
                    self._emit_connection_status(
                        ConnectionStatus.Active,
                        _("Spyder remote services are active"),
                    )
                    return True
                return False

            if not response.ok:
                try:
                    jupyter_error = await response.json()
                    self.logger.error(
                        "Unexpected jupyterhub response when starting "
                        "spyder's jupyter server: [%s]: %s",
                        jupyter_error["status"],
                        jupyter_error["message"],
                    )
                except (aiohttp.ContentTypeError, json.JSONDecodeError):
                    self.logger.exception(
                        "Unexpected jupyterhub response when starting "
                        "spyder's jupyter server: [%s]: %s",
                        response.status,
                        await response.text(),
                    )
                self._emit_connection_status(
                    ConnectionStatus.Error,
                    _("Error starting the remote server"),
                )
                return False

        async with self._session.get(
            f"hub/api/users/{self._user_name}/servers/spyder/progress",
        ) as response:
            if not response.ok:
                self.logger.error(
                    "Error getting remote server progress: %s",
                    response.status,
                )
                self._emit_connection_status(
                    ConnectionStatus.Error,
                    _(
                        "There was an error when trying to get the remote "
                        "server information"
                    ),
                )
                return False

            async for line in response.content:
                line = line.decode("utf8", "replace")
                if line.startswith("data:"):
                    ready = json.loads(line.split(":", 1)[1])["ready"]

        if not ready or not await self.check_server_version():
            self.logger.error(
                "Spyder remote server was unable to start, please check the "
                "JupyterHub logs for more information",
            )
            self._emit_connection_status(
                ConnectionStatus.Error, _("Error starting the remote server")
            )
            return False

        self._hb_task = asyncio.create_task(self._heartbeat())

        self._emit_connection_status(
            ConnectionStatus.Active, _("Spyder remote services are active")
        )
        return True

    async def ensure_server_installed(self) -> bool:
        """Assume the server is installed."""
        return True

    async def check_server_version(self) -> bool:
        """Check remote server version."""
        if not self.connected:
            self.logger.error("Connection is not open")
            self._emit_connection_status(
                ConnectionStatus.Error,
                _("JupyterHub is not connected"),
            )
            return False

        async with self._session.get(
            f"user/{self._user_name}/spyder/{SPYDER_PLUGIN_NAME}/version",
        ) as response:
            if not response.ok:
                try:
                    jupyter_error = await response.json()
                    self.logger.error(
                        "Unexpected jupyterhub response when getting "
                        "%s version: [%s]: %s",
                        SPYDER_PLUGIN_NAME,
                        jupyter_error["status"],
                        jupyter_error["message"],
                    )
                except (aiohttp.ContentTypeError, json.JSONDecodeError):
                    self.logger.exception(
                        "Error starting remote server: [%s]: %s",
                        response.status,
                        await response.text(),
                    )

                self._emit_connection_status(
                    ConnectionStatus.Error,
                    _("Error starting the remote server"),
                )
                return False

            version = await response.text()

        if Version(version) >= Version(SPYDER_REMOTE_MAX_VERSION):
            self.logger.error(
                "Server version mismatch: %s is greater than the maximum "
                "supported version %s",
                version,
                SPYDER_REMOTE_MAX_VERSION,
            )
            self._emit_version_mismatch(version)
            self._emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("Error staring the remote server"),
            )
            return False

        if Version(version) < Version(SPYDER_REMOTE_MIN_VERSION):
            self.logger.warning(
                "Server version mismatch: %s is lower than the minimum "
                "supported version %s. A more recent version will be "
                "installed.",
                version,
                SPYDER_REMOTE_MIN_VERSION,
            )
            return await self.install_remote_server()

        self.logger.info("Supported Server version: %s", version)

        return True

    async def _install_remote_server(self):
        self.logger.error(
            "Remote server installation is not supported for JupyterHub",
        )
        return False

    async def _create_new_connection(self) -> bool:
        """Create a new SSH connection."""
        self.logger.debug("Connecting to jupyterhub at %s", self.hub_url)

        if self._session is not None:
            await self._session.close()

        self._session = aiohttp.ClientSession(
            self.hub_url,
            headers={"Authorization": f"token {self.api_token}"},
        )

        user_data = None
        response = None
        try:
            async with self._session.get("hub/api/user") as response:
                if response.ok:
                    user_data = await response.json()
        except aiohttp.client_exceptions.ClientConnectionError as error:
            self.logger.error(
                "Error connecting to JupyterHub. The error was:<br>{}".format(
                    str(error)
                )
            )

        if user_data is None:
            if response is not None:
                self.logger.error(
                    "Error connecting to JupyterHub with code {}, which "
                    "corresponds to a {} reason".format(
                        response.status, response.reason
                    ),
                )

            # Close and reset sessions for discarded connections to allow
            # retries after connection info errors
            await self._session.close()
            self._session = None

            return False

        self._user_name = user_data["name"]

        self.logger.info("Successfully connected to JupyterHub")
        return True

    async def close_connection(self):
        """Close SSH connection."""
        if not self.connected:
            self.logger.debug("Connection is not open")
            return

        await self._session.close()
        self._session = None
        self.logger.info("Connection closed")
        self._reset_connection_established()
        self._emit_connection_status(
            ConnectionStatus.Inactive,
            _("The connection was closed successfully"),
        )

    async def stop_remote_server(self) -> bool:
        """Stop remote server."""
        if self._hb_task:
            self._hb_task.cancel()
            self._hb_task = None
        return await super().stop_remote_server()

    async def _heartbeat(self):
        while self.connected and self.server_started:
            await asyncio.sleep(30)
            try:
                async with self._session.get(
                    f"user/{self._user_name}/spyder/", timeout=30
                ) as response:
                    response.raise_for_status()
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                self.logger.warning("Heartbeat timed out")
                self._handle_connection_lost()
            except aiohttp.ClientConnectorError:
                self.logger.warning("Heartbeat connection error")
                self._handle_connection_lost()
            except Exception:
                self.logger.warning("Heartbeat failed", exc_info=True)
                self._handle_connection_lost()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Widget that handle communications between the IPython Console and
the Variable Explorer
"""

# Standard library imports
import logging
import os
from pickle import PicklingError, UnpicklingError
import sys

# Third-party imports
import cloudpickle
from packaging.version import parse
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from spyder_kernels.comms.commbase import CommError

# Local imports
from spyder.api.translations import _
from spyder.config.base import is_conda_based_app

# For logging
logger = logging.getLogger(__name__)

# Max time before giving up when making a blocking call to the kernel
CALL_KERNEL_TIMEOUT = 30

# URLs
GH_ISSUES = "https://github.com/spyder-ide/spyder/issues/new"
VAREXP_DONATIONS = (
    "https://www.spyder-ide.org/donate/variable-explorer-improvements"
)
PROJECTS_DOC_PAGE = "https://docs.spyder-ide.org/current/panes/projects.html"


class NamepaceBrowserWidget(RichJupyterWidget):
    """
    Widget with the necessary attributes and methods to handle communications
    between the IPython Console and the kernel namespace
    """
    # --- Public API --------------------------------------------------
    def get_value(self, name):
        """Ask kernel for a value"""
        # ---- Reasons
        reason_big = _("The variable is too big to be retrieved")
        reason_not_picklable = _(
            "It was not possible to create a copy of the variable in the "
            "kernel or to load that copy in Spyder.<br><br>"
            "If the object you're trying to view contains a generator, you "
            "need to replace it by a list because generators can't be "
            "serialized."
        )
        reason_dead = _("The kernel is dead")
        reason_other = _("An unknown error occurred, sorry.")
        reason_comm = _(
            "The channel used to communicate with the kernel is not working."
        )
        reason_missing_package_target = _(
            "The '<tt>{}</tt>' module is required to open this variable and "
            "it's not installed in the console environment. To fix this "
            "problem, please install it in the environment which you use to "
            "run your code."
        )
        reason_missing_package_installer = _(
            "The '<tt>{}</tt>' module is required to open this variable. "
            "Unfortunately, it's not part of our installer, which means your "
            "variable can't be displayed by Spyder."
        )
        reason_missing_package = _(
            "The '<tt>{}</tt>' module is required to open this variable and "
            "it's not installed alongside Spyder. To fix this problem, please "
            "install it in the same environment that you use to run Spyder."
        )
        reason_mismatched_numpy = _(
            "There is a mismatch between the Numpy versions used by Spyder "
            "and the kernel of your current console. To fix this problem, "
            "please upgrade <tt>numpy</tt> in the environment that you use to "
            "run Spyder to version 1.26.1 or higher."
        )
        reason_mismatched_pandas_2 = _(
            "There is a mismatch between the Pandas versions used by Spyder "
            "and the kernel of your current console. To fix this problem, "
            "please upgrade <tt>pandas</tt> in the console environment "
            "to version 2.0 or higher."
        )
        reason_mismatched_pandas_3 = _(
            "There is a mismatch between the Pandas versions used by Spyder "
            "and the kernel of your current console. To fix this problem, "
            "please upgrade <tt>pandas</tt> in the environment where Spyder "
            "is installed to version 3.0 or higher, or downgrade it in the "
            "console enviroment to version 2.3."
        )
        reason_mismatched_python_installer = _(
            "There is a mismatch between the Python versions used by Spyder "
            "({}) and the kernel of your current console ({}).<br><br>"
            "To fix it, you need to recreate your console environment with "
            "Python {} or {}."
        )
        reason_mismatched_python = _(
            "There is a mismatch between the Python versions used by Spyder "
            "({}) and the kernel of your current console ({}).<br><br>"
            "To fix it, you need to either install Spyder in an environment "
            "with Python {} or {}, or recreate your console environment with "
            "Python {} or {}."
        )

        # ---- Notes
        missing_package_note_1 = _(
            "If you want to see this fixed in the future, please make a "
            "donation <a href='{}'>here</a>."
        ).format(VAREXP_DONATIONS)

        # This is necessary to inform users what they need to do to explore
        # their own objects.
        # See spyder-ide/spyder#15988
        missing_package_note_2 = _(
            "If the required module is yours, you need to create a "
            "<a href='{}'>Spyder project</a> for it or add the path where "
            "it's located to the PYTHONPATH manager (available in the "
            "<tt>Tools</tt> menu)."
        ).format(PROJECTS_DOC_PAGE)

        # ---- Final message
        msg = _(
            "<br>%s<br><br>"
            "<b>Note</b>: If you consider this to be a valid error that needs "
            "to be fixed by the Spyder team, please report it on "
            "<a href='{}'>Github</a>."
        ).format(GH_ISSUES)
        msg_without_note = "<br>%s"

        # ---- Raise error which includes the message
        kernel_call_success = False
        show_full_msg = True
        try:
            value = self.call_kernel(
                blocking=True,
                # We prefer not to display errors because it's not clear that
                # they are related to what users are doing in the Variable
                # Explorer. So, it's not user friendly.
                # See spyder-ide/spyder#22411
                display_error=False,
                timeout=CALL_KERNEL_TIMEOUT
            ).get_value(name, encoded=True)
            kernel_call_success = True
            value = cloudpickle.loads(value)
            return value
        except TimeoutError:
            raise ValueError(msg % reason_big)
        except (PicklingError, UnpicklingError, TypeError) as err:
            if str(err).startswith(
                ("code expected at most", "code() argument")
            ):
                # This error happens when Spyder is using Python 3.11+ and the
                # kernel 3.10-, or the other way around. In that case,
                # cloudpickle can't deserialize the objects sent from the
                # kernel and we need to inform users about it.
                # Fixes spyder-ide/spyder#24125 and spyder-ide/spyder#24950.
                py_spyder_version = ".".join(
                    [str(n) for n in sys.version_info[:3]]
                )
                py_kernel_version = self.get_pythonenv_info()["python_version"]

                if parse(py_spyder_version) < parse(py_kernel_version):
                    (
                        py_spyder_good_version,
                        py_spyder_compatible_versions,
                        py_kernel_good_version,
                        py_kernel_compatible_versions,
                    ) = (
                        "3.11",
                        _("greater"),
                        "3.10",
                        _("lower"),
                    )
                else:
                    (
                        py_spyder_good_version,
                        py_spyder_compatible_versions,
                        py_kernel_good_version,
                        py_kernel_compatible_versions,
                    ) = (
                        "3.10",
                        _("lower"),
                        "3.11",
                        _("greater"),
                    )

                if is_conda_based_app():
                    raise ValueError(
                        msg
                        % reason_mismatched_python_installer.format(
                            py_spyder_version,
                            py_kernel_version,
                            py_kernel_good_version,
                            py_kernel_compatible_versions,
                        )
                    )
                else:
                    raise ValueError(
                        msg
                        % reason_mismatched_python.format(
                            py_spyder_version,
                            py_kernel_version,
                            py_spyder_good_version,
                            py_spyder_compatible_versions,
                            py_kernel_good_version,
                            py_kernel_compatible_versions,
                        )
                    )

            raise ValueError(msg % reason_not_picklable)
        except NotImplementedError as err:
            if "StringDtype(storage='python'" in str(err):
                raise ValueError(msg % reason_mismatched_pandas_3)
            else:
                raise ValueError(msg % reason_other)
        except RuntimeError:
            raise ValueError(msg % reason_dead)
        except KeyError:
            raise
        except CommError:
            raise ValueError(msg % reason_comm)
        except ModuleNotFoundError as e:
            if not kernel_call_success:
                name = e.args[0].error.name
                reason = reason_missing_package_target.format(name)
            elif e.name.startswith('numpy._core'):
                reason = reason_mismatched_numpy
            elif e.name == 'pandas.core.indexes.numeric':
                reason = reason_mismatched_pandas_2
            else:
                # We don't show the full message in this case so people don't
                # report this problem to Github and instead encourage them to
                # donate to the project that will solve the problem.
                # See spyder-ide/spyder#24922 for the details.
                show_full_msg = False

                if is_conda_based_app():
                    opening_paragraph = reason_missing_package_installer
                else:
                    opening_paragraph = reason_missing_package

                notes_vmargin = "0.4em" if os.name == "nt" else "0.3em"
                notes = (
                    "<style>"
                    "ul, li {{margin-left: -15px}}"
                    "li {{margin-bottom: {}}}"
                    "</style>"
                    "<ul>"
                    "<li>{}</li>"
                    "<li>{}</li>"
                    "</ul>"
                ).format(
                    notes_vmargin,
                    missing_package_note_1,
                    missing_package_note_2,
                )

                reason = (
                    opening_paragraph.format(e.name)
                    + "<br><br>"
                    + _("<b>Notes</b>:")
                    + notes
                )

            if show_full_msg:
                raise ValueError(msg % reason)
            else:
                raise ValueError(msg_without_note % reason)
        except Exception:
            raise ValueError(msg % reason_other)

    def set_value(self, name, value):
        """Set value for a variable"""
        reason_mismatched_numpy = _(
            "There is a mismatch between the Numpy versions used by Spyder "
            "and the kernel of your current console. To fix this problem, "
            "please upgrade <tt>numpy</tt> in the console environment to "
            "version 2.0 or higher."
        )
        msg = _(
            "<br>%s<br><br>"
            "<b>Note</b>: If you consider this to be a valid error that needs "
            "to be fixed by the Spyder team, please report it on "
            "<a href='{}'>Github</a>."
        ).format(GH_ISSUES)

        # Encode with cloudpickle and base64
        encoded_value = cloudpickle.dumps(value)

        try:
            self.call_kernel(
                interrupt=True,
                blocking=True,
                display_error=True,
            ).set_value(name, encoded_value, encoded=True)
        except ModuleNotFoundError as e:
            name = e.args[0].error.name
            if name.startswith('numpy._core'):
                raise ValueError(msg % reason_mismatched_numpy)
        except Exception:
            pass  # swallow exception

    def remove_value(self, name):
        """Remove a variable"""
        self.call_kernel(
            interrupt=True,
            blocking=False,
            display_error=True,
            ).remove_value(name)

    def copy_value(self, orig_name, new_name):
        """Copy a variable"""
        self.call_kernel(
            interrupt=True,
            blocking=False,
            display_error=True,
            ).copy_value(orig_name, new_name)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Connection status widget."""

from collections.abc import Iterable
import logging

import qstylizer.style
from qtpy.QtGui import QTextCursor
from qtpy.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.translations import _
from spyder.api.widgets.mixins import SvgToScaledPixmap
from spyder.plugins.remoteclient.api import MAX_CLIENT_MESSAGES
from spyder.plugins.remoteclient.api.protocol import (
    ConnectionInfo,
    ConnectionStatus,
    RemoteClientLog,
)
from spyder.plugins.remoteclient.widgets import AuthenticationMethod
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle, MAC
from spyder.widgets.simplecodeeditor import SimpleCodeEditor


# ---- Constants
# -----------------------------------------------------------------------------
STATUS_TO_TRANSLATION_STRINGS = {
    ConnectionStatus.Inactive: _("Inactive"),
    ConnectionStatus.Connecting: _("Connecting..."),
    ConnectionStatus.Connected: _("Connected..."),
    ConnectionStatus.Starting: _("Starting..."),
    ConnectionStatus.Active: _("Active"),
    ConnectionStatus.Stopping: _("Stopping..."),
    ConnectionStatus.Error: _("Error"),
}

STATUS_TO_COLOR = {
    ConnectionStatus.Inactive: SpyderPalette.COLOR_OCCURRENCE_5,
    ConnectionStatus.Connecting: SpyderPalette.COLOR_WARN_4,
    ConnectionStatus.Active: SpyderPalette.COLOR_SUCCESS_3,
    ConnectionStatus.Connected: SpyderPalette.COLOR_WARN_4,
    ConnectionStatus.Starting: SpyderPalette.COLOR_WARN_4,
    ConnectionStatus.Stopping: SpyderPalette.COLOR_WARN_4,
    ConnectionStatus.Error: SpyderPalette.COLOR_ERROR_2,
}

STATUS_TO_ICON = {
    ConnectionStatus.Inactive: "connection_disconnected",
    ConnectionStatus.Connecting: "connection_waiting",
    ConnectionStatus.Connected: "connection_waiting",
    ConnectionStatus.Starting: "connection_waiting",
    ConnectionStatus.Active: "connection_connected",
    ConnectionStatus.Stopping: "connection_waiting",
    ConnectionStatus.Error: "connection_error",
}

LOG_LEVEL_TO_FMT_STRING = {
    # It could be confusing to users to see a "Debug" message, so we prefer to
    # show it as "Info".
    logging.DEBUG: "<b>Info:</b>",
    logging.INFO: "<b>Info:</b>",
    logging.WARNING: (
        f'<span style="color:{STATUS_TO_COLOR[ConnectionStatus.Connecting]};">'
        f'<b>Warning:</b></span>'
    ),
    logging.ERROR: (
        f'<span style="color:{STATUS_TO_COLOR[ConnectionStatus.Error]};">'
        f'<b>Error:</b></span>'
    ),
    logging.CRITICAL: (
        f'<span style="color:{STATUS_TO_COLOR[ConnectionStatus.Error]};">'
        f'<b>Critical:</b></span>'
    ),
}


# ---- Widget
# -----------------------------------------------------------------------------
class ConnectionStatusWidget(
    QWidget,
    SpyderFontsMixin,
    SvgToScaledPixmap,
    SpyderConfigurationAccessor,
):

    CONF_SECTION = "remoteclient"

    def __init__(self, parent, host_id):
        super().__init__(parent)

        # Attributes
        self.host_id = host_id
        self.status = ConnectionStatus.Inactive

        # Widgets
        self._connection_label = QLabel(self)
        self._status_label = QLabel(self)
        self._user_label = QLabel(self)
        self._message_label = QLabel(self)
        self._message_label.setWordWrap(True)
        self._image_label = QLabel(self)
        self._log_label = QLabel(_("Connection messages"))
        self._log_widget = SimpleCodeEditor(self)
        self._copy_logs_button = QPushButton(_("Copy messages"))

        # Initial settings
        self._set_initial_text_in_labels()
        self._set_stylesheet()

        self._log_widget.setMaximumBlockCount(MAX_CLIENT_MESSAGES)
        self._log_widget.setReadOnly(True)
        self._log_widget.setMinimumHeight(210 if MAC else 230)
        self._log_widget.setPlaceholderText(_("No logs to show"))

        self._copy_logs_button.setEnabled(False)
        self._copy_logs_button.clicked.connect(self._copy_logs)

        # Info layout
        info_layout = QVBoxLayout()
        info_layout.setSpacing(0)
        info_layout.setContentsMargins(0, 0, 0, 0)
        info_layout.addWidget(self._connection_label)
        info_layout.addWidget(self._status_label)
        info_layout.addWidget(self._user_label)
        info_layout.addSpacing(4 * AppStyle.MarginSize)
        info_layout.addWidget(self._message_label)
        info_layout.addStretch()

        # This is necessary to align the image on the top side to the info
        # widgets to the left
        image_layout = QVBoxLayout()
        image_layout.setContentsMargins(0, 2 * AppStyle.MarginSize, 0, 0)
        image_layout.addWidget(self._image_label)

        # Top layout
        top_layout = QHBoxLayout()
        top_layout.addLayout(info_layout)
        top_layout.setStretchFactor(info_layout, 2)
        top_layout.addStretch()
        top_layout.addLayout(image_layout)

        # Bottom layout
        bottom_layout = QVBoxLayout()
        bottom_layout.setSpacing(0)
        bottom_layout.setContentsMargins(0, 0, 0, 0)
        bottom_layout.addWidget(self._log_label)
        bottom_layout.addWidget(self._log_widget)

        copy_layout = QHBoxLayout()
        copy_layout.addStretch()
        copy_layout.addWidget(self._copy_logs_button)
        bottom_layout.addSpacing(2 * AppStyle.MarginSize)
        bottom_layout.addLayout(copy_layout)

        # Final layout
        layout = QVBoxLayout()
        layout.setContentsMargins(
            # Match left margin of info widgets
            4 * AppStyle.MarginSize,
            0,
            # Match right margin of info widgets
            3 * AppStyle.MarginSize,
            # Prevent copy logs button to be clipped
            4 * AppStyle.MarginSize
        )
        layout.addLayout(top_layout)
        layout.addSpacing(6 * AppStyle.MarginSize)
        layout.addLayout(bottom_layout)
        self.setLayout(layout)

    # ---- Public API
    # -------------------------------------------------------------------------
    def update_status(self, info: ConnectionInfo):
        """Update graphical elements related to the connection status."""
        self.status = info["status"]
        message = info["message"]

        self._set_icon(self.status)
        self._set_text_in_labels(self.status)
        self._message_label.setText(message)

    def update_info(self):
        self._set_text_in_labels(self.status)

    def add_log(self, log: RemoteClientLog):
        """Add a new log message to the log widget."""
        if not self._copy_logs_button.isEnabled():
            self._copy_logs_button.setEnabled(True)

        formatted_log = (
            # Message
            f"<p>{LOG_LEVEL_TO_FMT_STRING[log['level']]} {log['message']}</p>"
            # Small vertical space to separate logs
            f"<div style='font-size: 2pt;font-weight: normal;'><p></p></div>"
        )

        # Move cursor so that new logs are always shown at the end
        self._log_widget.moveCursor(QTextCursor.End)
        self._log_widget.appendHtml(formatted_log)
        self._log_widget.moveCursor(QTextCursor.End)

    def add_logs(self, logs: Iterable):
        """Add saved log messages to the log widget."""
        for log in logs:
            self.add_log(log)

    # ---- Private API
    # -------------------------------------------------------------------------
    def _set_stylesheet(self):
        """Set stylesheet for elements in this widget."""
        # -- Style of important labels
        font_size = self.get_font(SpyderFontType.Interface).pointSize()
        important_labels_css = qstylizer.style.StyleSheet()
        important_labels_css.QLabel.setValues(
            fontSize=f"{font_size + 1}pt",
        )

        # Remove automatic indent added by Qt
        important_labels_css.setValues(**{'qproperty-indent': '0'})

        for label in [self._connection_label, self._message_label]:
            label.setStyleSheet(important_labels_css.toString())

        # -- Style of other info labels
        other_info_labels_css = qstylizer.style.StyleSheet()
        other_info_labels_css.setValues(
            marginLeft=f"{9 * AppStyle.MarginSize}px"
        )
        for label in [self._status_label, self._user_label]:
            label.setStyleSheet(other_info_labels_css.toString())

        # -- Style of log widgets
        log_label_css = qstylizer.style.StyleSheet()
        log_label_css.QLabel.setValues(
            # Increase padding (the default one is too small).
            padding=f"{2 * AppStyle.MarginSize}px",
            # Make it a bit different from a default QPushButton to not drag
            # the same amount of attention to it.
            backgroundColor=SpyderPalette.COLOR_BACKGROUND_3,
            # Remove bottom rounded borders
            borderBottomLeftRadius='0px',
            borderBottomRightRadius='0px',
            # This is necessary to align the label to the text above it
            marginLeft="2px",
        )
        self._log_label.setStyleSheet(log_label_css.toString())

        self._log_widget.css.QPlainTextEdit.setValues(
            # Remove these borders to make it appear attached to the top label
            borderTop="0px",
            borderTopLeftRadius='0px',
            borderTopRightRadius='0px',
            # Match border color with the top label one and avoid to change
            # that color when the widget is given focus
            borderLeft=f"1px solid {SpyderPalette.COLOR_BACKGROUND_3}",
            borderRight=f"1px solid {SpyderPalette.COLOR_BACKGROUND_3}",
            borderBottom=f"1px solid {SpyderPalette.COLOR_BACKGROUND_3}",
            # This is necessary to align the widget to the top label
            marginLeft="2px",
            # Increase padding a bit to make text look better
            paddingLeft="6px",
            paddingRight="6px",
            paddingTop="6px",
            # No need to have this due to the scrollbar
            paddingBottom="0px",
        )
        self._log_widget.setStyleSheet(self._log_widget.css.toString())

    def _set_initial_text_in_labels(self):
        status = self.get_conf(
            f"{self.host_id}/status", default=ConnectionStatus.Inactive
        )
        self._set_text_in_labels(status)
        self._set_icon(status)

        message = self.get_conf(f"{self.host_id}/status_message", default="")
        if not message:
            # This can only happen at startup or if the connection has never
            # been used
            message = _("The connection hasn't been used")
        self._message_label.setText(message)

    def _set_text_in_labels(self, status):
        color = STATUS_TO_COLOR[status]
        localized_status = STATUS_TO_TRANSLATION_STRINGS[status]

        address, username = self._get_address_and_username()

        self._connection_label.setText(
            _('Connection to: <span style="color:{}">{}<span>').format(
                color, address
            )
        )

        self._user_label.setText(
            _("Username: {}").format(username) if username else ""
        )

        self._status_label.setText(
            _('Status: <span style="color:{}">{}<span>').format(
                color, localized_status
            )
        )

    def _set_icon(self, status):
        self._image_label.setPixmap(
            self.svg_to_scaled_pixmap(STATUS_TO_ICON[status], rescale=1)
        )

    @property
    def _auth_method(self):
        """Get authentication method."""
        return self.get_conf(f"{self.host_id}/auth_method")

    def _copy_logs(self, clicked):
        """Copy log messages to clipboard."""
        text = self._log_widget.toPlainText()
        QApplication.clipboard().setText(text)

    def _get_address_and_username(self):
        if self._auth_method == AuthenticationMethod.JupyterHub:
            address = self.get_conf(f"{self.host_id}/url")
            username = ""
        # TODO: Address this for configfile login
        elif self._auth_method != AuthenticationMethod.ConfigFile:
            address = self.get_conf(
                f"{self.host_id}/{self._auth_method}/address"
            )
            username = self.get_conf(
                f"{self.host_id}/{self._auth_method}/username"
            )
        else:
            address = ""
            username = ""

        return (address, username)



# -----------------------------------------------------------------------------
# Copyright (c) 2024- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Status bar widget to show content from the IPython console.
"""

from __future__ import annotations

# Standard library imports
import functools
from typing import TYPE_CHECKING, Any

# Local imports
from spyder.api.shellconnect.mixins import ShellConnectMixin
from spyder.api.widgets.status import StatusBarWidget

if TYPE_CHECKING:
    import spyder.plugins.ipythonconsole.widgets
    from spyder.api.widgets.main_widget import PluginMainWidget


class ShellConnectStatusBarWidget(StatusBarWidget, ShellConnectMixin):
    """
    Base class for status bar widgets whose info depends on the current shell.
    """

    def __init__(self, parent: PluginMainWidget) -> None:
        """
        Create a new status bar widget to display content from a shell widget.

        Parameters
        ----------
        parent : PluginMainWidget
            The parent widget to this one.

        Returns
        -------
        None
        """
        super().__init__(parent)

        self.shellwidget_to_status: dict[
            spyder.plugins.ipythonconsole.widgets.ShellWidget, Any
        ] = {}
        """
        Mapping from shell widgets to their corresponding statuses.

        The specific type of status information stored will depend on
        what the implementing class needs to display.
        """

        self.current_shellwidget: (
            spyder.plugins.ipythonconsole.widgets.ShellWidget | None
        ) = None
        """Currently active associated shell widget."""

    # ---- Public API
    # -------------------------------------------------------------------------
    def update_status(self, status: Any):
        """
        Actions to take to update the widget status after switching consoles.

        Parameters
        ----------
        status : Any
            The shell widget status information needed by the status widget.
            To be defined by the implementing class.
        """
        raise NotImplementedError

    def on_kernel_start(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """Actions to take after the kernel starts.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget whose kernel was started.

        Returns
        -------
        None
        """
        raise NotImplementedError

    def config_spyder_kernel(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """Actions to take to configure the kernel.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget whose kernel will be configured.

        Returns
        -------
        None
        """
        pass

    # ---- ShellConnectMixin API
    # -------------------------------------------------------------------------
    def set_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """
        Actions to take when setting (i.e. giving focus to) a shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that will be given focus.

        Returns
        -------
        None
        """
        self.current_shellwidget = None
        if shellwidget in self.shellwidget_to_status:
            status = self.shellwidget_to_status[shellwidget]
            if status is None:
                # This means the console failed to start or there was a problem
                # obtaining the status from the kernel. So, we don't need to
                # show this widget in those cases.
                self.hide()
            else:
                self.show()
                self.current_shellwidget = shellwidget
                self.update_status(status)

    def add_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """Actions to take when adding a shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that will be added.

        Returns
        -------
        None
        """
        shellwidget.sig_config_spyder_kernel.connect(
            functools.partial(self.config_spyder_kernel, shellwidget)
        )
        shellwidget.sig_kernel_is_ready.connect(
            # We can't use functools.partial here because it gives memory leaks
            # in Python versions older than 3.10
            lambda sw=shellwidget: self.on_kernel_start(sw)
        )

        self.shellwidget_to_status[shellwidget] = None
        self.set_shellwidget(shellwidget)

    def remove_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """Actions to take when removing a shell widget.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that will have its associated widget removed.

        Returns
        -------
        None
        """
        if shellwidget in self.shellwidget_to_status:
            self.shellwidget_to_status.pop(shellwidget)

    def add_errored_shellwidget(
        self,
        shellwidget: spyder.plugins.ipythonconsole.widgets.ShellWidget,
    ) -> None:
        """Actions to take when a shell widget has failed with an error.

        Parameters
        ----------
        shellwidget : spyder.plugins.ipythonconsole.widgets.ShellWidget
            The shell widget that failed with an error.

        Returns
        -------
        None
        """
        self.shellwidget_to_status[shellwidget] = None
        self.set_shellwidget(shellwidget)



# -----------------------------------------------------------------------------
# Copyright (c) 2016- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Base classes, mixins and widgets for creating plugins to extend Spyder.

This API should be considered production-ready as of Spyder 6.0.
The API version is modified according to the following rules:

* If a module/class/method/function/signal is added, then the minor version
  will be incremented.

* If a module/class/method/function/signal is removed, renamed or changes its
  signature, then the major version will be incremented.

* Whenever possible, deprecation warnings and alerts will be employed to
  inform developers of breaking-compatibility changes in a future API release
  In such cases, the minor version will be increased and once the deprecated
  APIs are removed then the major version will be updated.
"""

VERSION_INFO: tuple[int, int, int] = (2, 0, 0)
"""Tuple form of API version, broken down into ``(major, minor, micro)``."""

__version__: str = str(".".join(map(str, VERSION_INFO)))
"""
Spyder API version; minor bumped for additions/deprecations, major for breaks.

The minor version will be incremented when adding or deprecating an API,
(for example, a module, class, method, function, signal, constant, or similar)
while the major version will be incremented for all breaking API changes,
such as removing, renaming or changing the signature of an API.
"""



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Run dialogs and widgets and data models."""

# Standard library imports
import os.path as osp
import textwrap
from typing import Optional, Tuple, List, Dict
from uuid import uuid4

# Third party imports
from qtpy.compat import getexistingdirectory
from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import (
    QAction,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QLayout,
    QMessageBox,
    QRadioButton,
    QPushButton,
    QStackedWidget,
    QVBoxLayout,
)
import qstylizer.style

# Local imports
from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.translations import _
from spyder.api.widgets.comboboxes import SpyderComboBox
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.config.base import running_in_ci, running_under_pytest
from spyder.plugins.run.api import (
    ExtendedRunExecutionParameters,
    RunExecutorConfigurationGroup,
    RunExecutionParameters,
    SupportedExecutionRunConfiguration,
    WorkingDirOpts,
    WorkingDirSource,
)
from spyder.utils.icon_manager import ima
from spyder.utils.misc import getcwd_or_home
from spyder.utils.qthelpers import qapplication
from spyder.utils.stylesheet import AppStyle, MAC
from spyder.widgets.collapsible import CollapsibleWidget
from spyder.widgets.helperwidgets import TipWidget


# ---- Main constants
# -----------------------------------------------------------------------------
FILE_DIR = _("The directory of the file being executed")
CW_DIR = _("The current working directory")
FIXED_DIR = _("The following directory:")
EMPTY_NAME = _("Provide a name for this configuration")
REPEATED_NAME = _("Select a different name for this configuration")
SAME_PARAMETERS = _(
    "You are trying to save a configuration that is exactly the same as the "
    "current one"
)


class RunDialogStatus:
    Close = 0
    Save = 1
    Run = 2


# ---- Base class
# -----------------------------------------------------------------------------
class BaseRunConfigDialog(QDialog):
    """Run configuration dialog box, base widget"""
    size_change = Signal(QSize)

    def __init__(self, parent=None, disable_run_btn=False):
        QDialog.__init__(self, parent)
        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        # Destroying the C++ object right after closing the dialog box,
        # otherwise it may be garbage-collected in another QThread
        # (e.g. the editor's analysis thread in Spyder), thus leading to
        # a segmentation fault on UNIX or an application crash on Windows
        self.setAttribute(Qt.WA_DeleteOnClose)

        self.setWindowIcon(ima.icon('run_settings'))
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.disable_run_btn = disable_run_btn

        # Style that will be set by children
        self._css = qstylizer.style.StyleSheet()

    def add_widgets(self, *widgets_or_spacings):
        """Add widgets/spacing to dialog vertical layout"""
        layout = self.layout()

        for widget_or_spacing in widgets_or_spacings:
            if isinstance(widget_or_spacing, int):
                layout.addSpacing(widget_or_spacing)
            elif isinstance(widget_or_spacing, QLayout):
                layout.addLayout(widget_or_spacing)
            else:
                layout.addWidget(widget_or_spacing)

        return layout

    def add_button_box(self, stdbtns):
        """Create dialog button box and add it to the dialog layout"""
        self.bbox = SpyderDialogButtonBox(stdbtns)

        if not self.disable_run_btn:
            run_btn = self.bbox.addButton(
                _("Run"), QDialogButtonBox.ActionRole)
            run_btn.clicked.connect(self.run_btn_clicked)

        reset_deafults_btn = self.bbox.addButton(
            _('Reset'), QDialogButtonBox.ResetRole)
        reset_deafults_btn.clicked.connect(self.reset_btn_clicked)

        # Align this button to the text above it
        reset_deafults_btn.setStyleSheet("margin-left: 5px")

        self.bbox.accepted.connect(self.accept)
        self.bbox.rejected.connect(self.reject)

        self.layout().addWidget(self.bbox)

    def resizeEvent(self, event):
        """
        Reimplement Qt method to be able to save the widget's size from the
        main application
        """
        QDialog.resizeEvent(self, event)
        self.size_change.emit(self.size())

    def run_btn_clicked(self):
        """Run button was just clicked"""
        pass

    def reset_btn_clicked(self):
        """Reset button was clicked."""
        pass

    def ok_btn_clicked(self):
        """Ok button was clicked."""
        pass

    def setup(self):
        """Setup Run Configuration dialog with filename *fname*"""
        raise NotImplementedError


# ---- Dialogs
# -----------------------------------------------------------------------------
class ExecutionParametersDialog(BaseRunConfigDialog):
    """Run execution parameters edition dialog."""

    def __init__(
        self,
        parent,
        executor_name,
        executor_params: Dict[
            Tuple[str, str], SupportedExecutionRunConfiguration
        ],
        param_names: Dict[Tuple[str, str], List[str]],
        extensions: Optional[List[str]] = None,
        contexts: Optional[Dict[str, List[str]]] = None,
        current_params: Optional[ExtendedRunExecutionParameters] = None,
        extension: Optional[str] = None,
        context: Optional[str] = None,
        new_config: bool = False
    ):
        super().__init__(parent, True)

        self.executor_name = executor_name
        self.executor_params = executor_params
        self.param_names = param_names
        self.current_params = current_params
        self.extensions = extensions or []
        self.contexts = contexts or {}
        self.extension = extension
        self.context = context
        self.new_config = new_config

        self.parameters_name = None
        if current_params is not None:
            self.parameters_name = (
                _("Default")
                if current_params["default"]
                else current_params["name"]
            )

        self.current_widget = None
        self.status = RunDialogStatus.Close
        self.saved_conf = None

    # ---- Public methods
    # -------------------------------------------------------------------------
    def setup(self):
        # --- Configuration name
        if self.new_config:
            params_name_text = _("Save configuration as:")
        else:
            params_name_text = _("Configuration name:")

        params_name_label = QLabel(params_name_text)
        self.store_params_text = QLineEdit(self)
        self.store_params_text.setMinimumWidth(250)
        store_params_layout = QHBoxLayout()
        store_params_layout.addWidget(params_name_label)
        store_params_layout.addWidget(self.store_params_text)
        store_params_layout.addStretch(1)

        # This action needs to be added before setting an icon for it so that
        # it doesn't show up in the line edit (despite being set as not visible
        # below). That's probably a Qt bug.
        status_action = QAction(self)
        self.store_params_text.addAction(
            status_action, QLineEdit.TrailingPosition
        )
        self.store_params_text.status_action = status_action

        status_action.setIcon(ima.icon("error"))
        status_action.setVisible(False)

        # This is necessary to fix the style of the tooltip shown inside the
        # lineedit
        store_params_css = qstylizer.style.StyleSheet()
        store_params_css["QLineEdit QToolTip"].setValues(
            padding="1px 2px",
        )
        self.store_params_text.setStyleSheet(store_params_css.toString())

        # --- Extension and context widgets
        ext_combo_label = QLabel(_("File extension:"))
        context_combo_label = QLabel(_("Run context:"))

        self.extension_combo = SpyderComboBox(self)
        self.extension_combo.addItems(self.extensions)
        self.extension_combo.currentIndexChanged.connect(
            self.extension_changed)

        self.context_combo = SpyderComboBox(self)
        self.context_combo.currentIndexChanged.connect(self.context_changed)

        self.extension_combo.setMinimumWidth(150)
        self.context_combo.setMinimumWidth(150)

        ext_context_g_layout = QGridLayout()
        ext_context_g_layout.addWidget(ext_combo_label, 0, 0)
        ext_context_g_layout.addWidget(self.extension_combo, 0, 1)
        ext_context_g_layout.addWidget(context_combo_label, 1, 0)
        ext_context_g_layout.addWidget(self.context_combo, 1, 1)

        ext_context_layout = QHBoxLayout()
        ext_context_layout.addLayout(ext_context_g_layout)
        ext_context_layout.addStretch(1)

        # --- Runner settings
        self.stack = QStackedWidget(self)

        # --- Working directory settings
        self.wdir_group = QGroupBox(_("Working directory settings"))
        wdir_layout = QVBoxLayout(self.wdir_group)
        wdir_layout.setContentsMargins(
            3 * AppStyle.MarginSize,
            3 * AppStyle.MarginSize,
            3 * AppStyle.MarginSize,
            AppStyle.MarginSize if MAC else 0,
        )

        self.file_dir_radio = QRadioButton(FILE_DIR)
        wdir_layout.addWidget(self.file_dir_radio)

        self.cwd_radio = QRadioButton(CW_DIR)
        wdir_layout.addWidget(self.cwd_radio)

        self.fixed_dir_radio = QRadioButton(FIXED_DIR)
        self.wd_edit = QLineEdit(self)
        self.fixed_dir_radio.toggled.connect(self.wd_edit.setEnabled)
        self.wd_edit.setEnabled(False)
        browse_btn = QPushButton(ima.icon('DirOpenIcon'), '', self)
        browse_btn.setToolTip(_("Select directory"))
        browse_btn.clicked.connect(self.select_directory)
        browse_btn.setIconSize(
            QSize(AppStyle.ConfigPageIconSize, AppStyle.ConfigPageIconSize)
        )

        fixed_dir_layout = QHBoxLayout()
        fixed_dir_layout.addWidget(self.fixed_dir_radio)
        fixed_dir_layout.addWidget(self.wd_edit)
        fixed_dir_layout.addWidget(browse_btn)
        wdir_layout.addLayout(fixed_dir_layout)

        # --- Final layout
        layout = self.add_widgets(
            store_params_layout,
            4 * AppStyle.MarginSize,
            ext_context_layout,
            (3 if MAC else 4) * AppStyle.MarginSize,
            self.stack,
            self.wdir_group,
            (-2 if MAC else 1) * AppStyle.MarginSize,
        )
        layout.addStretch()
        layout.setContentsMargins(*((AppStyle.InnerContentPadding,) * 4))

        self.add_button_box(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        # --- Settings
        self.setWindowTitle(
            _("New run configuration for: {}").format(self.executor_name)
        )
        self.layout().setSizeConstraint(QLayout.SetFixedSize)

        extension_index = 0
        if self.extension is not None:
            extension_index = self.extensions.index(self.extension)
            self.extension_combo.setEnabled(False)

        self.extension_combo.setCurrentIndex(extension_index)

        # This is necessary because extension_changed is not triggered
        # automatically for this extension_index.
        if extension_index == 0:
            self.extension_changed(extension_index)

        if self.context is not None:
            self.context_combo.setEnabled(False)

        if self.parameters_name:
            self.store_params_text.setText(self.parameters_name)

            # Don't allow to change name for default or already saved params.
            if self.current_params["default"] or not self.new_config:
                self.store_params_text.setEnabled(False)

        # --- Stylesheet
        self.setStyleSheet(self._stylesheet)

    def extension_changed(self, index: int):
        if index < 0:
            return

        self.selected_extension = self.extension_combo.itemText(index)
        contexts = self.contexts[self.selected_extension]

        self.context_combo.clear()
        self.context_combo.addItems(contexts)
        self.context_combo.setCurrentIndex(-1)

        context_index = 0
        if self.context is not None:
            context_index = contexts.index(self.context)
        self.context_combo.setCurrentIndex(context_index)

    def context_changed(self, index: int, reset: bool = False):
        if index < 0:
            return

        # Clear the QStackWidget contents
        self.current_widget = None
        while self.stack.count() > 0:
            widget = self.stack.widget(0)
            self.stack.removeWidget(widget)

        self.selected_context = self.context_combo.itemText(index)

        executor_conf_metadata = self.executor_params[
            (self.selected_extension, self.selected_context)]

        requires_cwd = executor_conf_metadata['requires_cwd']
        ConfigWidget = (executor_conf_metadata['configuration_widget'] or
                        RunExecutorConfigurationGroup)

        if executor_conf_metadata['configuration_widget'] is None:
            self.stack.setEnabled(False)
        else:
            self.stack.setEnabled(True)

        self.wdir_group.setEnabled(requires_cwd)

        self.current_widget = ConfigWidget(
            self, self.selected_context, self.selected_extension, {})
        self.stack.addWidget(self.current_widget)

        working_dir_params = WorkingDirOpts(
            source=WorkingDirSource.ConfigurationDirectory,
            path=None)
        exec_params = RunExecutionParameters(
            working_dir=working_dir_params,
            executor_params=None)

        default_params = self.current_widget.get_default_configuration()

        if self.current_params:
            params = self.current_params['params']
            working_dir_params = params['working_dir']
            exec_params = params

        params_set = (
            default_params
            if reset
            else (exec_params["executor_params"] or default_params)
        )

        if params_set.keys() == default_params.keys():
            self.current_widget.set_configuration(params_set)

        source = working_dir_params['source']
        path = working_dir_params['path']

        if source == WorkingDirSource.ConfigurationDirectory:
            self.file_dir_radio.setChecked(True)
            self.cwd_radio.setChecked(False)
            self.fixed_dir_radio.setChecked(False)
            self.wd_edit.setText('')
        elif source == WorkingDirSource.CurrentDirectory:
            self.file_dir_radio.setChecked(False)
            self.cwd_radio.setChecked(True)
            self.fixed_dir_radio.setChecked(False)
            self.wd_edit.setText('')
        elif source == WorkingDirSource.CustomDirectory:
            self.file_dir_radio.setChecked(False)
            self.cwd_radio.setChecked(False)
            self.fixed_dir_radio.setChecked(True)
            self.wd_edit.setText(path)

        if not self.stack.isEnabled() and not self.wdir_group.isEnabled():
            ok_btn = self.bbox.button(QDialogButtonBox.Ok)
            ok_btn.setEnabled(False)

        self.adjustSize()

    def select_directory(self):
        """Select directory"""
        basedir = str(self.wd_edit.text())
        if not osp.isdir(basedir):
            basedir = getcwd_or_home()
        directory = getexistingdirectory(self, _("Select directory"), basedir)
        if directory:
            self.wd_edit.setText(directory)
            self.dir = directory

    def reset_btn_clicked(self):
        index = self.context_combo.currentIndex()
        self.context_changed(index, reset=True)

    def run_btn_clicked(self):
        self.status |= RunDialogStatus.Run

    def ok_btn_clicked(self):
        self.status |= RunDialogStatus.Save

    def get_configuration(
            self
    ) -> Tuple[str, str, ExtendedRunExecutionParameters]:

        return self.saved_conf

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def accept(self) -> None:
        self.status |= RunDialogStatus.Save
        widget_conf = self.current_widget.get_configuration()
        self.store_params_text.status_action.setVisible(False)

        path = None
        source = None
        if self.file_dir_radio.isChecked():
            source = WorkingDirSource.ConfigurationDirectory
        elif self.cwd_radio.isChecked():
            source = WorkingDirSource.CurrentDirectory
        else:
            source = WorkingDirSource.CustomDirectory
            path = self.wd_edit.text()

        cwd_opts = WorkingDirOpts(source=source, path=path)

        exec_params = RunExecutionParameters(
            working_dir=cwd_opts, executor_params=widget_conf)

        if self.current_params:
            uuid = self.current_params['uuid']
        else:
            uuid = str(uuid4())

        # Validate name only for new configurations
        name = self.store_params_text.text()
        if self.new_config:
            if name == '':
                self.store_params_text.status_action.setVisible(True)
                self.store_params_text.status_action.setToolTip(
                    '\n'.join(textwrap.wrap(EMPTY_NAME, 50))
                )
                return
            else:
                extension = self.extension_combo.lineEdit().text()
                context = self.context_combo.lineEdit().text()
                current_names = self.param_names[(extension, context)]
                if name in current_names:
                    self.store_params_text.status_action.setVisible(True)
                    self.store_params_text.status_action.setToolTip(
                        '\n'.join(textwrap.wrap(REPEATED_NAME, 50))
                    )
                    return

        # Check if params are app default ones.
        # Fixes spyder-ide/spyder#22649
        if self.current_params is None:
            # The user is trying to create new params, so this is not a
            # default.
            is_default = False
        else:
            if self.current_params["default"]:
                # Default params
                is_default = True
            else:
                # User created params
                is_default = False

        ext_exec_params = ExtendedRunExecutionParameters(
            uuid=uuid,
            name=name,
            params=exec_params,
            file_uuid=None,
            default=is_default,
        )

        self.saved_conf = (self.selected_extension, self.selected_context,
                           ext_exec_params)

        super().accept()

    # ---- Private methods
    # -------------------------------------------------------------------------
    @property
    def _stylesheet(self):
        # This avoids the extra bottom margin added by the config dialog since
        # this widget is one of its children
        self._css.QGroupBox.setValues(
            marginBottom='0px',
        )

        return self._css.toString()


class RunDialog(BaseRunConfigDialog, SpyderFontsMixin):
    """Run dialog used to configure run executors."""

    sig_delete_config_requested = Signal(str, str, str, str)

    def __init__(
        self,
        parent=None,
        run_conf_model=None,
        executors_model=None,
        parameter_model=None,
        disable_run_btn=False
    ):
        super().__init__(parent, disable_run_btn=disable_run_btn)

        self.run_conf_model = run_conf_model
        self.executors_model = executors_model
        self.parameter_model = parameter_model

        self.current_widget = None
        self.status = RunDialogStatus.Close
        self._is_shown = False

    # ---- Public methods
    # -------------------------------------------------------------------------
    def setup(self):
        # --- Header
        self.header_label = QLabel(self)
        self.header_label.setObjectName("run-header-label")

        # --- File combobox
        # It's hidden by default to decrease the complexity of this dialog
        self.configuration_combo = SpyderComboBox(self)
        self.configuration_combo.hide()

        # --- Executor and parameters widgets
        executor_label = QLabel(_("Runner:"))
        self.executor_combo = SpyderComboBox(self)
        self.executor_combo.setMinimumWidth(250)
        executor_tip = TipWidget(
            _("Configure the selected runner for this file"),
            icon=ima.icon('question_tip'),
            hover_icon=ima.icon('question_tip_hover'),
            size=23,
            wrap_text=True
        )

        parameters_label = QLabel(_("Preset configuration:"))
        self.parameters_combo = SpyderComboBox(self)
        self.parameters_combo.setMinimumWidth(250)
        parameters_tip = TipWidget(
            _(
                "Select between global or local (i.e. for this file) "
                "configuration presets. You can set the latter below"
            ),
            icon=ima.icon('question_tip'),
            hover_icon=ima.icon('question_tip_hover'),
            size=23,
            wrap_text=True
        )

        executor_g_layout = QGridLayout()
        executor_g_layout.addWidget(executor_label, 0, 0)
        executor_g_layout.addWidget(self.executor_combo, 0, 1)
        executor_g_layout.addWidget(executor_tip, 0, 2)
        executor_g_layout.addWidget(parameters_label, 1, 0)
        executor_g_layout.addWidget(self.parameters_combo, 1, 1)
        executor_g_layout.addWidget(parameters_tip, 1, 2)

        executor_layout = QHBoxLayout()
        executor_layout.addLayout(executor_g_layout)
        executor_layout.addStretch()

        # --- Configuration properties
        config_props_group = QGroupBox(_("Configuration properties"))
        config_props_layout = QGridLayout(config_props_group)

        # Increase margin between title and line edit below so this looks good
        config_props_margins = config_props_layout.contentsMargins()
        config_props_margins.setTop(12)
        config_props_layout.setContentsMargins(config_props_margins)

        # Name to save custom configuration
        name_params_label = QLabel(_("Name:"))
        self.name_params_text = QLineEdit(self)
        self.name_params_text.setPlaceholderText(
            _("Set a name for this configuration")
        )
        name_params_tip = TipWidget(
            _(
                "You can set as many configuration presets as you want by "
                "providing different names. Each one will be saved after "
                "clicking the Ok button below"
            ),
            icon=ima.icon('question_tip'),
            hover_icon=ima.icon('question_tip_hover'),
            size=23,
            wrap_text=True
        )

        # This action needs to be added before setting an icon for it so that
        # it doesn't show up in the line edit (despite being set as not visible
        # below). That's probably a Qt bug.
        status_action = QAction(self)
        self.name_params_text.addAction(
            status_action, QLineEdit.TrailingPosition
        )
        self.name_params_text.status_action = status_action

        status_action.setIcon(ima.icon("error"))
        status_action.setVisible(False)

        config_props_layout.addWidget(name_params_label, 0, 0)
        config_props_layout.addWidget(self.name_params_text, 0, 1)
        config_props_layout.addWidget(name_params_tip, 0, 2)

        # --- Runner settings
        self.stack = QStackedWidget(self)

        # --- Working directory settings
        self.wdir_group = QGroupBox(_("Working directory settings"))
        wdir_layout = QVBoxLayout(self.wdir_group)

        self.file_dir_radio = QRadioButton(FILE_DIR)
        wdir_layout.addWidget(self.file_dir_radio)

        self.cwd_radio = QRadioButton(CW_DIR)
        wdir_layout.addWidget(self.cwd_radio)

        self.fixed_dir_radio = QRadioButton(FIXED_DIR)
        self.wd_edit = QLineEdit(self)
        self.fixed_dir_radio.toggled.connect(self.wd_edit.setEnabled)
        self.wd_edit.setEnabled(False)
        browse_btn = QPushButton(ima.icon('DirOpenIcon'), '', self)
        browse_btn.setToolTip(_("Select directory"))
        browse_btn.clicked.connect(self.select_directory)
        browse_btn.setIconSize(
            QSize(AppStyle.ConfigPageIconSize, AppStyle.ConfigPageIconSize)
        )

        fixed_dir_layout = QHBoxLayout()
        fixed_dir_layout.addWidget(self.fixed_dir_radio)
        fixed_dir_layout.addWidget(self.wd_edit)
        fixed_dir_layout.addWidget(browse_btn)
        wdir_layout.addLayout(fixed_dir_layout)

        # --- Group all customization widgets into a collapsible one
        custom_config = CollapsibleWidget(self, _("Custom configuration"))
        custom_config.addWidget(config_props_group)
        custom_config.addWidget(self.stack)
        custom_config.addWidget(self.wdir_group)

        # Fix bottom and left margins.
        custom_config.set_content_bottom_margin(0)
        custom_config.set_content_right_margin(AppStyle.MarginSize)

        # Center dialog after custom_config is expanded/collapsed
        custom_config._animation.finished.connect(self._center_dialog)

        # --- Final layout
        layout = self.add_widgets(
            self.header_label,
            self.configuration_combo,  # Hidden for simplicity
            executor_layout,
            custom_config,
            (-2 if MAC else 1) * AppStyle.MarginSize,
        )
        layout.setContentsMargins(
            AppStyle.InnerContentPadding,
            # This needs to be bigger to make the layout look better
            AppStyle.InnerContentPadding + AppStyle.MarginSize,
            # This makes the left and right padding be the same
            AppStyle.InnerContentPadding + 4,
            AppStyle.InnerContentPadding,
        )

        self.add_button_box(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.delete_button = QPushButton(_("Delete"))
        self.delete_button.clicked.connect(self.delete_btn_clicked)
        self.bbox.addButton(self.delete_button, QDialogButtonBox.ActionRole)

        # --- Settings
        self.executor_combo.currentIndexChanged.connect(
            self.display_executor_configuration)
        self.executor_combo.setModel(self.executors_model)

        # This signal needs to be connected after
        # executor_combo.currentIndexChanged and before
        # configuration_combo.currentIndexChanged for parameters_combo to be
        # updated as expected when opening the dialog.
        self.parameters_combo.currentIndexChanged.connect(
            self.update_parameter_set
        )
        self.parameters_combo.setModel(self.parameter_model)

        self.configuration_combo.currentIndexChanged.connect(
            self.update_configuration_run_index)
        self.configuration_combo.setModel(self.run_conf_model)
        self.configuration_combo.setCurrentIndex(
            self.run_conf_model.get_initial_index())
        self.configuration_combo.setMaxVisibleItems(1)

        self.executor_combo.setMaxVisibleItems(20)
        self.executor_combo.view().setVerticalScrollBarPolicy(
            Qt.ScrollBarAsNeeded)

        self.setWindowTitle(_("Run configuration per file"))
        self.layout().setSizeConstraint(QLayout.SetFixedSize)

        self.setStyleSheet(self._stylesheet)

    def select_directory(self):
        """Select directory"""
        basedir = str(self.wd_edit.text())
        if not osp.isdir(basedir):
            basedir = getcwd_or_home()
        directory = getexistingdirectory(self, _("Select directory"), basedir)
        if directory:
            self.wd_edit.setText(directory)
            self.dir = directory

    def update_configuration_run_index(self, index: int):
        self.executor_combo.setCurrentIndex(-1)
        self.run_conf_model.update_index(index)
        self.executor_combo.setCurrentIndex(
            self.executors_model.get_initial_index()
        )

    def update_parameter_set(self, index: int):
        if index < 0:
            return

        # Get parameters
        stored_params = self.parameter_model.get_parameters(index)
        global_params = stored_params["file_uuid"] is None

        # Set parameters name
        if global_params:
            # We set this name for global params so users don't have to think
            # about selecting one when customizing them
            custom_name = self._get_auto_custom_name(stored_params["name"])
            self.name_params_text.setText(custom_name)
        else:
            # We show the actual name for file params
            self.name_params_text.setText(stored_params["name"])

        # Disable delete button for global configs
        if global_params:
            self.delete_button.setEnabled(False)
        else:
            self.delete_button.setEnabled(True)

        # Set parameters in their corresponding graphical elements
        params = stored_params["params"]
        working_dir_params = params['working_dir']
        exec_params = params['executor_params']
        self.current_widget.set_configuration(exec_params)

        source = working_dir_params['source']
        path = working_dir_params['path']

        if source == WorkingDirSource.ConfigurationDirectory:
            self.file_dir_radio.setChecked(True)
            self.cwd_radio.setChecked(False)
            self.fixed_dir_radio.setChecked(False)
            self.wd_edit.setText('')
        elif source == WorkingDirSource.CurrentDirectory:
            self.file_dir_radio.setChecked(False)
            self.cwd_radio.setChecked(True)
            self.fixed_dir_radio.setChecked(False)
            self.wd_edit.setText('')
        elif source == WorkingDirSource.CustomDirectory:
            self.file_dir_radio.setChecked(False)
            self.cwd_radio.setChecked(False)
            self.fixed_dir_radio.setChecked(True)
            self.wd_edit.setText(path)

    def display_executor_configuration(self, index: int):
        if index == -1:
            return

        # Clear the QStackWidget contents
        self.current_widget = None
        while self.stack.count() > 0:
            widget = self.stack.widget(0)
            self.stack.removeWidget(widget)

        exec_tuple = self.executors_model.get_selected_run_executor(index)
        executor_name, executor_info = exec_tuple
        enable_cwd = executor_info['requires_cwd']
        self.wdir_group.setEnabled(enable_cwd)

        ConfigWidget = (executor_info['configuration_widget'] or
                        RunExecutorConfigurationGroup)

        if executor_info['configuration_widget'] is None:
            self.stack.setVisible(False)
        else:
            self.stack.setVisible(True)

        metadata = self.run_conf_model.get_selected_metadata()
        context = metadata['context']
        input_extension = metadata['input_extension']
        uuid = metadata['uuid']

        self.current_widget = ConfigWidget(
            self, context, input_extension, metadata)
        self.stack.addWidget(self.current_widget)

        if uuid not in self.run_conf_model:
            return

        stored_params = self.run_conf_model.get_run_configuration_parameters(
            uuid, executor_name)['params']

        # Only show global parameters (i.e. those with file_uuid = None) or
        # those that correspond to the current file.
        stored_params = {
            k: v for (k, v) in stored_params.items()
            if v.get("file_uuid") in [None, uuid]
        }

        self.parameter_model.set_parameters(stored_params)
        selected_params = self.run_conf_model.get_last_used_execution_params(
            uuid, executor_name)
        params_index = self.parameter_model.get_parameters_index_by_uuid(
            selected_params
        )

        self.parameters_combo.setCurrentIndex(params_index)
        self.adjustSize()

    def select_executor(self, executor_name: str):
        self.executor_combo.setCurrentIndex(
            self.executors_model.get_run_executor_index(executor_name))

    def reset_btn_clicked(self):
        self.parameters_combo.setCurrentIndex(0)

    def run_btn_clicked(self):
        self.status |= RunDialogStatus.Run
        self.accept()

    def delete_btn_clicked(self):
        answer = QMessageBox.question(
            self,
            _("Delete"),
            _("Do you want to delete the current configuration?"),
        )

        if answer == QMessageBox.Yes:
            # Get executor name
            executor_name, __ = self.executors_model.get_selected_run_executor(
                self.executor_combo.currentIndex()
            )

            # Get extension and context_id
            metadata = self.run_conf_model.get_selected_metadata()
            extension = metadata["input_extension"]
            context_id = metadata["context"]["identifier"]

            # Get index associated with config
            idx = self.parameters_combo.currentIndex()

            # Get config uuid
            uuid, __ = self.parameter_model.get_parameters_uuid_name(idx)

            self.sig_delete_config_requested.emit(
                executor_name, extension, context_id, uuid
            )

            # Close dialog to not have to deal with the difficult case of
            # updating its contents after this config is deleted
            self.reject()

    def get_configuration(
        self
    ) -> Tuple[str, str, ExtendedRunExecutionParameters, bool]:

        return self.saved_conf

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def accept(self) -> None:
        self.status |= RunDialogStatus.Save

        # Configuration to save/execute
        widget_conf = self.current_widget.get_configuration()

        # Hide status action in case users fix the problem reported through it
        # on a successive try
        self.name_params_text.status_action.setVisible(False)

        # Get index of current params
        current_index = self.parameters_combo.currentIndex()
        if running_in_ci() and current_index == -1:
            # This error seems to happen only on CIs
            self.status = RunDialogStatus.Close
            return

        # Detect if the current params are global
        params = self.parameter_model.get_parameters(current_index)
        global_params = params["file_uuid"] is None

        if global_params:
            custom_name = self._get_auto_custom_name(params["name"])
        else:
            custom_name = ""

        # Working directory params
        path = None
        source = None
        if self.file_dir_radio.isChecked():
            source = WorkingDirSource.ConfigurationDirectory
        elif self.cwd_radio.isChecked():
            source = WorkingDirSource.CurrentDirectory
        else:
            source = WorkingDirSource.CustomDirectory
            path = self.wd_edit.text()

        cwd_opts = WorkingDirOpts(source=source, path=path)

        # Execution params
        exec_params = RunExecutionParameters(
            working_dir=cwd_opts, executor_params=widget_conf
        )

        # Different validations for the params name
        params_name = self.name_params_text.text()
        if self.isVisible():
            allow_to_close = True

            if not params_name:
                # Don't allow to save params without a name
                self.name_params_text.status_action.setVisible(True)
                self.name_params_text.status_action.setToolTip(
                    '\n'.join(textwrap.wrap(EMPTY_NAME, 50))
                )
                allow_to_close = False
            elif global_params and params_name == custom_name:
                # We don't need to perform a validation in this case because we
                # set the params name on behalf of users
                pass
            elif params_name != self.parameters_combo.lineEdit().text():
                if params_name in self.parameter_model.get_parameter_names():
                    # Don't allow to save params with the same name of an
                    # existing one because it doesn't make sense.
                    allow_to_close = False
                    self.name_params_text.status_action.setVisible(True)
                    self.name_params_text.status_action.setToolTip(
                        '\n'.join(textwrap.wrap(REPEATED_NAME, 50))
                    )
                elif params["params"] == exec_params:
                    # Don't allow to save params that are exactly the same as
                    # the current ones.
                    allow_to_close = False
                    self.name_params_text.status_action.setVisible(True)
                    self.name_params_text.status_action.setToolTip(
                       '\n'.join(textwrap.wrap(SAME_PARAMETERS, 50))
                    )

            if not allow_to_close:
                # With this the dialog can be closed when clicking the Cancel
                # button
                self.status = RunDialogStatus.Close
                return

        # Get index associated with config
        if params["params"] == exec_params:
            # This avoids saving an unnecessary custom config when the current
            # parameters haven't been modified with respect to the selected
            # config
            idx = current_index
        else:
            idx = self.parameter_model.get_parameters_index_by_name(
                params_name
            )

        # Get uuid and name from index
        if idx == -1:
            # This means that there are no saved parameters for params_name, so
            # we need to generate a new uuid for them.
            uuid = str(uuid4())
            name = params_name
        else:
            # Retrieve uuid and name from our config system
            uuid, name = self.parameter_model.get_parameters_uuid_name(idx)

        # Build configuration to be saved or executed
        metadata_info = self.run_conf_model.get_metadata(
            self.configuration_combo.currentIndex()
        )

        ext_exec_params = ExtendedRunExecutionParameters(
            uuid=uuid,
            name=name,
            params=exec_params,
            file_uuid=None
            if (global_params and idx >= 0)
            else metadata_info["uuid"],
            default=True
            if (global_params and params["default"] and idx >= 0)
            else False,
        )

        executor_name, __ = self.executors_model.get_selected_run_executor(
            self.executor_combo.currentIndex()
        )

        self.saved_conf = (
            metadata_info["uuid"],
            executor_name,
            ext_exec_params,
        )

        return super().accept()

    def showEvent(self, event):
        """Adjustments when the widget is shown."""
        if not self._is_shown:
            # Set file name as the header
            fname = self.configuration_combo.currentText()
            header_font = (
                self.get_font(SpyderFontType.Interface, font_size_delta=1)
            )

            # Elide fname in case fname is too long
            fm = QFontMetrics(header_font)
            text = fm.elidedText(
                fname, Qt.ElideLeft, self.header_label.width()
            )

            self.header_label.setFont(header_font)
            self.header_label.setAlignment(Qt.AlignCenter)
            self.header_label.setText(text)
            if text != fname:
                self.header_label.setToolTip(fname)

            self._is_shown = True

        super().showEvent(event)

    # ---- Private methods
    # -------------------------------------------------------------------------
    @property
    def _stylesheet(self):
        # --- Style for the header
        self._css["QLabel#run-header-label"].setValues(
            # Add good enough margin with the widgets below it.
            marginBottom=f"{3 * AppStyle.MarginSize}px",
            # This is necessary to align the label to the widgets below it.
            marginLeft="4px",
        )

        # --- Style for the collapsible
        self._css["CollapsibleWidget"].setValues(
            # Separate it from the widgets above it
            marginTop=f"{3 * AppStyle.MarginSize}px"
        )

        return self._css.toString()

    def _center_dialog(self):
        """
        Center dialog relative to the main window after collapsing/expanding
        the custom configuration widget.
        """
        # This doesn't work in our tests because the main window is usually
        # not available in them.
        if running_under_pytest():
            return

        qapp = qapplication()
        main_window_pos = qapp.get_mainwindow_position()
        main_window_height = qapp.get_mainwindow_height()

        # We only center the dialog vertically because there's no need to
        # do it horizontally.
        x = self.x()
        y = main_window_pos.y() + ((main_window_height - self.height()) // 2)

        self.move(x, y)

    def _get_auto_custom_name(self, global_params_name: str) -> str:
        """
        Get an auto-generated custom name given the a global parameters one.
        """
        n_custom = self.parameter_model.get_number_of_custom_params(
            global_params_name
        )

        return (
            global_params_name
            + " ("
            + _("custom")
            + (")" if n_custom == 0 else f" {n_custom})")
        )



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Switcher Proxy Model."""

# Third party imports
from qtpy.QtCore import QSortFilterProxyModel, Qt


class SwitcherProxyModel(QSortFilterProxyModel):
    """A proxy model to perform sorting on the scored items."""

    def __init__(self, parent=None):
        """Proxy model to perform sorting on the scored items."""
        super().__init__(parent)
        self.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.setSortCaseSensitivity(Qt.CaseInsensitive)
        self.setDynamicSortFilter(True)
        self.__filter_by_score = False

    def set_filter_by_score(self, value):
        """
        Set whether the items should be filtered by their score result.

        Parameters
        ----------
        value : bool
           Indicates whether the items should be filtered by their
           score result.
        """
        self.__filter_by_score = value
        self.invalidateFilter()

    def filterAcceptsRow(self, source_row, source_parent):
        """Override Qt method to filter items by their score result."""
        item = self.sourceModel().item(source_row)
        if self.__filter_by_score is False or item.is_action_item():
            return True
        else:
            return not item.get_score() == -1

    def sortBy(self, attr):
        """Override Qt method."""
        self.__sort_by = attr
        self.invalidate()
        self.sort(0, Qt.AscendingOrder)

    def lessThan(self, left, right):
        """Override Qt method."""
        left_item = self.sourceModel().itemFromIndex(left)
        right_item = self.sourceModel().itemFromIndex(right)

        # Check for attribute, otherwise, check for data
        if (
            hasattr(left_item, self.__sort_by)
            and getattr(left_item, self.__sort_by) is not None
            and hasattr(right_item, self.__sort_by)
            and getattr(right_item, self.__sort_by) is not None
        ):
            left_data = getattr(left_item, self.__sort_by)
            right_data = getattr(right_item, self.__sort_by)
            return left_data < right_data

        return super().lessThan(left, right)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# based on pylintgui.py by Pierre Raybaut
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Profiler widget.

See the official documentation on python profiling:
https://docs.python.org/3/library/profile.html
"""

# Standard library imports
import functools
import os.path as osp

# Third party imports
from qtpy.compat import getopenfilename, getsavefilename
from qtpy.QtCore import Signal
from superqt.utils import signals_blocked

# Local imports
from spyder.api.translations import _
from spyder.api.shellconnect.main_widget import ShellConnectMainWidget
from spyder.plugins.profiler.widgets.profiler_data_tree import (
    ProfilerSubWidget
)
from spyder.utils.icon_manager import ima
from spyder.utils.misc import getcwd_or_home


class ProfilerWidgetActions:
    # Triggers
    Clear = 'clear_action'
    Collapse = 'collapse_action'
    Expand = 'expand_action'
    CallersOrCallees = "callers_or_callees_action"
    ToggleBuiltins = "toggle_builtins_action"
    Home = "HomeAction"
    SlowLocal = 'slow_local_action'
    LoadData = 'load_data_action'
    SaveData = 'save_data_action'
    Search = "find_action"
    Undo = "undo_action"
    Redo = "redo_action"
    Stop = "stop_action"


class ProfilerWidgetMenus:
    EmptyContextMenu = 'empty'
    PopulatedContextMenu = 'populated'


class ProfilerContextMenuSections:
    Locals = 'locals_section'
    Other = "other_section"


class ProfilerWidgetContextMenuActions:
    GotoDefinition = "goto_definition_action"
    ShowCallees = "show_callees_action"
    ShowCallers = "show_callers_action"


class ProfilerWidgetMainToolbarSections:
    # BrowseView = "view_section" # To be added later
    ExpandCollapse = "collapse_section"
    ChangeView = "change_view_section"
    Stop = "stop_section"


# --- Widgets
# ----------------------------------------------------------------------------
class ProfilerWidget(ShellConnectMainWidget):
    """Profiler widget."""

    # PluginMainWidget API
    ENABLE_SPINNER = True
    SHOW_MESSAGE_WHEN_EMPTY = True
    IMAGE_WHEN_EMPTY = "code-profiler"
    MESSAGE_WHEN_EMPTY = _("Code not profiled yet")
    DESCRIPTION_WHEN_EMPTY = _(
        "Profile your code to explore which functions and methods took the "
        "longest to run and were called the most, and find out where to "
        "optimize it."
    )

    # Other
    TIP_CALLERS = _("Show functions or modules that call the root item")
    TIP_CALLEES = _("Show functions or modules called by the root item")
    TIP_CALLERS_OR_CALLEES = _(
        "Show functions or modules that call an item or are called by it"
    )

    # --- Signals
    # ------------------------------------------------------------------------
    sig_edit_goto_requested = Signal(str, int, str)
    """
    This signal will request to open a file in a given row and column
    using a code editor.

    Parameters
    ----------
    path: str
        Path to file.
    row: int
        Cursor starting row position.
    word: str
        Word to select on given row.
    """

    def __init__(self, name=None, plugin=None, parent=None):
        super().__init__(name, plugin, parent)

    # ---- PluginMainWidget API
    # -------------------------------------------------------------------------
    def get_title(self):
        return _('Profiler')

    def setup(self):
        # ---- Toolbar actions
        collapse_action = self.create_action(
            ProfilerWidgetActions.Collapse,
            text=_('Collapse'),
            tip=_('Collapse one level up'),
            icon=self.create_icon('collapse'),
            triggered=self._collapse_tree,
        )
        expand_action = self.create_action(
            ProfilerWidgetActions.Expand,
            text=_('Expand'),
            tip=_('Expand one level down'),
            icon=self.create_icon('expand'),
            triggered=self._expand_tree,
        )
        callers_or_callees_action = self.create_action(
            ProfilerWidgetActions.CallersOrCallees,
            text=self.TIP_CALLERS_OR_CALLEES,
            tip=self.TIP_CALLERS_OR_CALLEES,
            icon=self.create_icon("callers_or_callees"),
            toggled=self._toggle_callers_or_callees,
        )
        slow_local_action = self.create_action(
            ProfilerWidgetActions.SlowLocal,
            text=_("Show items with large local time"),
            tip=_('Show items with large local time'),
            icon=self.create_icon('slow'),
            toggled=self._slow_local_tree,
        )
        toggle_builtins_action = self.create_action(
            ProfilerWidgetActions.ToggleBuiltins,
            text=_("Hide calls to external libraries"),
            tip=_('Hide calls to external libraries'),
            icon=self.create_icon('hide'),
            toggled=self._toggle_builtins,
        )
        save_action = self.create_action(
            ProfilerWidgetActions.SaveData,
            text=_("Save data"),
            tip=_('Save profiling data'),
            icon=self.create_icon('filesave'),
            triggered=self._save_data,
        )
        load_action = self.create_action(
            ProfilerWidgetActions.LoadData,
            text=_("Load data"),
            tip=_('Load profiling data for comparison'),
            icon=self.create_icon('fileimport'),
            triggered=self._load_data,
        )
        clear_action = self.create_action(
            ProfilerWidgetActions.Clear,
            text=_("Clear comparison"),
            tip=_("Clear comparison"),
            icon=self.create_icon('editdelete'),
            triggered=self._clear,
        )
        search_action = self.create_action(
            ProfilerWidgetActions.Search,
            text=_("Search"),
            icon=self.create_icon('find'),
            toggled=self._toggle_finder,
            register_shortcut=True
        )
        stop_action = self.create_action(
            ProfilerWidgetActions.Stop,
            text=_("Stop profiling"),
            icon=self.create_icon('stop_profile'),
            triggered=self._stop_profiling,
        )

        # This needs to be workedd out better because right now is confusing
        # and kind of unnecessary
        # undo_action = self.create_action(
        #     ProfilerWidgetActions.Undo,
        #     text=_("Previous View"),
        #     icon=self.create_icon('previous'),
        #     triggered=self._undo,
        #     register_shortcut=True
        # )
        # redo_action = self.create_action(
        #     ProfilerWidgetActions.Redo,
        #     text=_("Next View"),
        #     icon=self.create_icon('next'),
        #     triggered=self._redo,
        #     register_shortcut=True
        # )
        # home_action = self.create_action(
        #     ProfilerWidgetActions.Home,
        #     text=_("Reset tree"),
        #     tip=_('Go back to full tree'),
        #     icon=self.create_icon('home'),
        #     triggered=self._home_tree,
        # )

        # ---- Main Toolbar
        main_toolbar = self.get_main_toolbar()

        # To be added later
        # for action in [undo_action, redo_action, home_action]:
        #     self.add_item_to_toolbar(
        #         action,
        #         toolbar=main_toolbar,
        #         section=ProfilerWidgetMainToolbarSections.BrowseView,
        #     )

        for action in [collapse_action, expand_action]:
            self.add_item_to_toolbar(
                action,
                toolbar=main_toolbar,
                section=ProfilerWidgetMainToolbarSections.ExpandCollapse,
            )

        for action in [
            slow_local_action,
            toggle_builtins_action,
            callers_or_callees_action,
            search_action
        ]:
            self.add_item_to_toolbar(
                action,
                toolbar=main_toolbar,
                section=ProfilerWidgetMainToolbarSections.ChangeView,
            )

        self.add_item_to_toolbar(
            stop_action,
            toolbar=main_toolbar,
            section=ProfilerWidgetMainToolbarSections.Stop,
        )

        # ---- Corner widget
        for action in [save_action, load_action, clear_action]:
            self.add_corner_widget(action, before=self._options_button)

        # ---- Context menu actions
        show_callees_action = self.create_action(
            ProfilerWidgetContextMenuActions.ShowCallees,
            _("Show functions or modules called by this item"),
            icon=self.create_icon('callees'),
            triggered=self._show_callees
        )
        show_callers_action = self.create_action(
            ProfilerWidgetContextMenuActions.ShowCallers,
            _("Show functions or modules that call this item"),
            icon=self.create_icon('callers'),
            triggered=self._show_callers
        )
        goto_definition_action = self.create_action(
            ProfilerWidgetContextMenuActions.GotoDefinition,
            _("Go to definition"),
            icon=self.create_icon("transparent"),
            triggered=self._goto_definition
        )

        self._context_menu = self.create_menu(
            ProfilerWidgetMenus.PopulatedContextMenu
        )
        for item in [show_callers_action, show_callees_action]:
            self.add_item_to_menu(
                item,
                menu=self._context_menu,
                section=ProfilerContextMenuSections.Locals,
            )

        self.add_item_to_menu(
            goto_definition_action,
            menu=self._context_menu,
            section=ProfilerContextMenuSections.Other,
        )

    def update_actions(self):
        """Update actions."""
        widget = self.current_widget()
        search_action = self.get_action(ProfilerWidgetActions.Search)
        callers_or_callees_action = self.get_action(
            ProfilerWidgetActions.CallersOrCallees
        )
        toggle_builtins_action = self.get_action(
            ProfilerWidgetActions.ToggleBuiltins
        )
        slow_local_action = self.get_action(ProfilerWidgetActions.SlowLocal)
        stop_action = self.get_action(ProfilerWidgetActions.Stop)

        widget_inactive = (
            widget is None or self.is_current_widget_error_message()
        )
        if widget_inactive:
            search = False
            callers_or_callees_enabled = False
            ignore_builtins = False
            show_slow = False
            stop = False
            self.stop_spinner()
        else:
            search = widget.finder_is_visible()
            callers_or_callees_enabled = widget.callers_or_callees_enabled
            ignore_builtins = widget.ignore_builtins
            show_slow = widget.show_slow
            stop = widget.is_profiling

        toggle_builtins_action.setChecked(ignore_builtins)
        stop_action.setEnabled(stop)

        # Showing callers/callees can't be combined with slow locals and search
        # because they give different views, so we need to disable them.
        if callers_or_callees_enabled:
            # Automatically toggle the action
            callers_or_callees_action.setEnabled(True)
            callers_or_callees_action.setChecked(True)

            # This prevents an additional call to update_actions because
            # refresh_tree emits at the end sig_refresh
            with signals_blocked(widget):
                widget.refresh_tree()

            # Adjust button's tooltip and icon
            show_callers = widget.inverted_tree
            callers_or_callees_action.setToolTip(
                self.TIP_CALLERS if show_callers else self.TIP_CALLEES
            )
            callers_or_callees_action.setIcon(
                ima.icon("callers" if show_callers else "callees")
            )

            # Disable slow locals
            widget.show_slow = False
            with signals_blocked(slow_local_action):
                slow_local_action.setChecked(False)
                slow_local_action.setEnabled(False)

            # Disable search and hide finder widget
            with signals_blocked(search_action):
                search_action.setChecked(False)
                search_action.setEnabled(False)

            with signals_blocked(widget.finder):
                widget.finder.set_visible(False)

            # We expand the tree so that users can easily inspect callers or
            # callees
            self._expand_tree()
        else:
            search_action.setChecked(search)
            slow_local_action.setChecked(show_slow)

            # Recreate custom view if we're showing slow locals or searching
            # for something
            if not widget_inactive and widget.recreate_custom_view:
                # Reset state for next time
                widget.recreate_custom_view = False

                if search:
                    search_text = widget.finder_text()
                    if search_text:
                        widget.do_find(search_text)
                elif show_slow:
                    self._slow_local_tree(True)

            if callers_or_callees_action.isEnabled():
                # Change icon and tooltip when the button is inactive
                callers_or_callees_action.setToolTip(
                    self.TIP_CALLERS_OR_CALLEES
                )
                callers_or_callees_action.setIcon(
                    ima.icon("callers_or_callees")
                )

                # Untoggle the button
                with signals_blocked(callers_or_callees_action):
                    callers_or_callees_action.setChecked(False)
                    callers_or_callees_action.setEnabled(False)

        if not widget_inactive:
            if widget.is_profiling:
                self.start_spinner()
            else:
                self.stop_spinner()

        # Home, undo and redo are disabled for now because they are confusing
        # and kind of unnecessary
        # can_redo = False
        # can_undo = False

        tree_empty = True
        can_clear = False
        if not widget_inactive:
            tree_empty = widget.profdata is None
            # can_undo = len(widget.data_tree.history) > 1
            # can_redo = len(widget.data_tree.redo_history) > 0
            can_clear = widget.compare_data is not None

        for action_name in [
            ProfilerWidgetActions.Collapse,
            ProfilerWidgetActions.Expand,
            ProfilerWidgetActions.ToggleBuiltins,
            # ProfilerWidgetActions.Home,
            ProfilerWidgetActions.SlowLocal,
            ProfilerWidgetActions.SaveData,
            ProfilerWidgetActions.LoadData,
            ProfilerWidgetActions.Search,
        ]:
            action = self.get_action(action_name)
            if action_name in [
                ProfilerWidgetActions.SlowLocal,
                ProfilerWidgetActions.Search,
            ]:
                action.setEnabled(
                    not tree_empty and not callers_or_callees_enabled
                )
            elif action_name == ProfilerWidgetActions.LoadData:
                action.setEnabled(not widget_inactive)
            else:
                action.setEnabled(not tree_empty)

        # undo_action = self.get_action(ProfilerWidgetActions.Undo)
        # redo_action = self.get_action(ProfilerWidgetActions.Redo)

        # undo_action.setEnabled(can_undo)
        # redo_action.setEnabled(can_redo)

        clear_action = self.get_action(ProfilerWidgetActions.Clear)
        clear_action.setEnabled(can_clear)

    # ---- ShellConnectPluginMixin API
    # -------------------------------------------------------------------------
    def create_new_widget(self, shellwidget):
        """Create new profiler widget."""
        widget = ProfilerSubWidget(self)
        widget.sig_display_requested.connect(self._display_request)
        widget.sig_refresh.connect(self.update_actions)
        widget.set_context_menu(self._context_menu)
        widget.sig_hide_finder_requested.connect(self._hide_finder)
        widget.sig_show_empty_message_requested.connect(
            self.switch_empty_message
        )

        shellwidget.register_kernel_call_handler(
            "show_profile_file", widget.show_profile_buffer
        )
        shellwidget.register_kernel_call_handler(
            "start_profiling", self._start_profiling
        )
        widget.on_kernel_ready_callback = functools.partial(
            self._on_kernel_ready, widget
        )
        shellwidget.sig_kernel_is_ready.connect(
            widget.on_kernel_ready_callback
        )

        widget.shellwidget = shellwidget
        return widget

    def close_widget(self, widget):
        """Close profiler widget."""
        widget.sig_refresh.disconnect(self.update_actions)
        widget.sig_display_requested.disconnect(self._display_request)
        widget.sig_hide_finder_requested.disconnect(self._hide_finder)

        # Unregister
        widget.shellwidget.unregister_kernel_call_handler("show_profile_file")
        widget.shellwidget.unregister_kernel_call_handler("start_profiling")
        widget.shellwidget.sig_kernel_is_ready.disconnect(
            widget.on_kernel_ready_callback
        )
        widget.setParent(None)
        widget.close()

    def switch_widget(self, widget, old_widget):
        """Switch widget."""
        pass

    def switch_empty_message(self, value: bool):
        """
        Override this method to prevent hiding the empty message if profiling
        finishes in another console but the current one has no content to show.
        """
        widget = self.current_widget()
        if widget is None:
            return

        if value:
            self.show_empty_message()
        else:
            if widget.profdata is not None:
                self.show_content_widget()

    def current_widget(self) -> ProfilerSubWidget:
        """Override to add typing."""
        return super().current_widget()

    # ---- Private API
    # -------------------------------------------------------------------------
    def _start_profiling(self):
        self.start_spinner()

        stop_action = self.get_action(ProfilerWidgetActions.Stop)
        stop_action.setEnabled(True)

        widget = self.current_widget()
        if widget is None:
            return

        widget.is_profiling = True

        # Check if we're showing slow locals or searching for something to
        # recreate the custom view after new results arrive.
        if widget.show_slow or (
            widget.finder_is_visible() and widget.finder_text()
        ):
            widget.recreate_custom_view = True

        # Reset state of callers/callees view because the new results couldn't
        # contain the selected item.
        widget.callers_or_callees_enabled = False
        widget.inverted_tree = False

    def _stop_profiling(self):
        widget = self.current_widget()
        if widget is None or self.is_current_widget_error_message():
            return

        if widget.is_profiling:
            self.stop_spinner()

            stop_action = self.get_action(ProfilerWidgetActions.Stop)
            stop_action.setEnabled(False)

            widget.shellwidget.request_interrupt_kernel()
            widget.is_profiling = False

    def _home_tree(self):
        """Show home tree."""
        widget = self.current_widget()
        if widget is None:
            return
        widget.home_tree()

    def _toggle_callers_or_callees(self, state):
        """
        Toggle filter for callers or callees.

        Notes
        -----
        * The toogle state is handled automatically by update_actions.
        * After users untoggle the button, they'll return to the initial view.
        """
        widget = self.current_widget()
        if widget is None:
            return

        if not state:
            widget.callers_or_callees_enabled = False
            widget.inverted_tree = False
            self._home_tree()

    def _collapse_tree(self):
        self.current_widget().change_view(-1)

    def _expand_tree(self):
        self.current_widget().change_view(1)

    def _toggle_builtins(self, state):
        """Toggle builtins."""
        widget = self.current_widget()
        if widget is None:
            return
        widget.ignore_builtins = state
        widget.refresh_tree()

    def _slow_local_tree(self, state):
        """Show items with large local times"""
        widget = self.current_widget()
        if widget is None:
            return

        widget.show_slow = state
        if state:
            widget.show_slow_items()
        else:
            self._home_tree()

    def _undo(self):
        """Undo change."""
        widget = self.current_widget()
        if widget is None:
            return
        widget.undo()

    def _redo(self):
        """Redo changes."""
        widget = self.current_widget()
        if widget is None:
            return
        widget.redo()

    def _show_callers(self):
        """Show callers."""
        widget = self.current_widget()
        if widget is None:
            return

        widget.inverted_tree = True
        widget.show_selected()

    def _show_callees(self):
        """Show callees."""
        widget = self.current_widget()
        if widget is None:
            return

        widget.inverted_tree = False
        widget.show_selected()

    def _goto_definition(self):
        widget = self.current_widget()
        if widget is None:
            return

        item = widget.currentItem()
        if osp.isfile(item.filename):
            self.sig_edit_goto_requested.emit(
                item.filename, item.line_number, ""
            )

    def _save_data(self):
        """Save data."""
        widget = self.current_widget()
        if widget is None:
            return
        title = _("Save profiler result")
        filename, _selfilter = getsavefilename(
            self,
            title,
            getcwd_or_home(),
            _("Profiler result") + " (*.prof)",
        )
        extension = osp.splitext(filename)[1].lower()
        if not extension:
            # Needed to prevent trying to save a data file without extension
            # See spyder-ide/spyder#19633
            filename = filename + '.prof'

        if filename:
            widget.save_data(filename)

    def _load_data(self):
        """Compare previous saved run with last run."""
        widget = self.current_widget()
        if widget is None:
            return
        filename, _selfilter = getopenfilename(
            self,
            _("Select script to compare"),
            getcwd_or_home(),
            _("Profiler result") + " (*.prof)",
        )

        if filename:
            widget.compare(filename)
            widget.home_tree()
            self.update_actions()

    def _clear(self):
        """Clear data in tree."""
        widget = self.current_widget()
        if widget is None:
            return
        widget.compare(None)
        widget.home_tree()
        self.update_actions()

    def _display_request(self, widget):
        """
        Display request from ProfilerDataTree.

        Only display if this is the current widget.
        """
        self.update_actions()
        self._stop_profiling()

        if (
            self.current_widget() is widget
            and self.get_conf("switch_to_plugin")
        ):
            self.get_plugin().switch_to_plugin()

    def _toggle_finder(self, checked):
        """Show or hide finder."""
        widget = self.current_widget()
        if widget is None:
            return
        widget.toggle_finder(checked)

    def _hide_finder(self):
        """Hide finder."""
        action = self.get_action(ProfilerWidgetActions.Search)
        action.setChecked(False)

    def _on_kernel_ready(self, widget: ProfilerSubWidget):
        self._stop_profiling()
        widget.set_pane_empty(True)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Clipboard helper module.
"""
from qtpy.QtWidgets import QApplication


class ClipboardHelper:
    # Clipboard metadata
    metadata_hash = None
    metadata_indent = None
    metadata_tab_stop_width_spaces = None

    def get_current_hash(self):
        clipboard = QApplication.clipboard()
        return hash(str(clipboard.text()))

    def get_line_indentation(self, text, tab_stop_width_spaces=None):
        """Get indentation for given line."""
        if tab_stop_width_spaces:
            text = text.replace("\t", " " * tab_stop_width_spaces)
        return len(text)-len(text.lstrip())

    def save_indentation(self, preceding_text, tab_stop_width_spaces=None):
        """
        Save the indentation corresponding to the clipboard data.

        Must be called right after copying.
        """
        self.metadata_hash = self.get_current_hash()
        self.metadata_indent = self.get_line_indentation(
            preceding_text, tab_stop_width_spaces)
        self.metadata_tab_stop_width_spaces = tab_stop_width_spaces

    def remaining_lines_adjustment(self, preceding_text):
        """
        Get remaining lines adjustments needed to keep multiline
        pasted text consistant.
        """
        if self.get_current_hash() == self.metadata_hash:
            return (
                self.get_line_indentation(
                    preceding_text,
                    self.metadata_tab_stop_width_spaces)
                - self.metadata_indent)
        return 0


CLIPBOARD_HELPER = ClipboardHelper()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Profiler config page."""

from qtpy.QtWidgets import QVBoxLayout

from spyder.api.preferences import PluginConfigPage
from spyder.api.translations import _


class ProfilerConfigPage(PluginConfigPage):
    def setup_page(self):
        switch_to_plugin_cb = self.create_checkbox(
            _("Open profiler when profiling finishes"),
            "switch_to_plugin",
            tip=_(
                "This option switches to the profiler plugin "
                "when a profiling has ended."
            ),
        )

        slow_spin = self.create_spinbox(
            _("Maximum number of items displayed with large local time"),
            "",
            'n_slow_children',
            min_=1,
            max_=1000,
            step=1
        )

        vlayout = QVBoxLayout()
        vlayout.addWidget(switch_to_plugin_cb)
        vlayout.addWidget(slow_spin)
        vlayout.addStretch(1)
        self.setLayout(vlayout)
